/*
 * $Id: arch-linux.c,v 1.3 1998/07/15 09:31:47 fengmd Exp $
 */
/*
 * $Log: arch-linux.c,v $
 * Revision 1.3  1998/07/15 09:31:47  fengmd
 * Support user-inlet and library, merge changes from the main branch.
 *
 * Revision 1.2  1998/06/01 14:32:45  fengmd
 * Merge changes from the main chunk.
 *
 * Revision 1.1.1.1  1998/03/19 04:43:35  astark
 * Initial creation of umbrella/all-sets implementation
 *
 * Revision 1.1.1.1.2.2  1997/06/16 16:00:20  fengmd
 * Remerge from the base for Version 5.0.1.
 *
 * Revision 1.5  1997/06/16 14:31:17  athena
 * Minor changes for linux.  Updated FAQ.
 *
 * Revision 1.4  1997/06/09 19:03:03  athena
 * Changed RTS so that Cilk won't dump core by default
 *
 * Revision 1.3  1997/03/02 21:06:23  randall
 * Made sure Cilk_die doesn't return.
 *
 * Revision 1.2  1997/03/01 17:42:28  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <stdarg.h>
#include <stdio.h>
#include <cilk.h>
#include <cilk-internal.h>
#include <cilk-lib.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/types.h>
#define MAP_FAILED ((void *) -1)

/***********************************************************\
 * Debugging & tracing
\***********************************************************/

DECLARE_SHARED_VARIABLE(static Cilk_mutex, dprintf_lock, );

void Cilk_dprintf(const char *fmt,...)
{
     va_list l;

     /* ensure nice interleaving of debugging messages */
     Cilk_mutex_wait(&USE_SHARED(dprintf_lock));
     if (Self >= 0)
	  fprintf(stderr, "%d: ", Self);
     else
	  fprintf(stderr, "global: ");
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     fflush(stderr);
     va_end(l);
     Cilk_mutex_signal(&USE_SHARED(dprintf_lock));
}

DECLARE_SHARED_VARIABLE(static Cilk_mutex, die_lock, );

void Cilk_die(const char *fmt,...)
{
     va_list l;

     Cilk_mutex_wait(&USE_SHARED(dprintf_lock));
     if (Self >= 0)
	  fprintf(stderr, "%d: fatal error: ", Self);
     else
	  fprintf(stderr, "global: fatal error: ");
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     va_end(l);
     Cilk_mutex_signal(&USE_SHARED(dprintf_lock));

     /* ensure only the first fatal error produces death of the program */
     if (USE_SHARED_RO(dump_core) && Cilk_mutex_try(&USE_SHARED(die_lock)))
	  abort();		/* dump core */
     else
          exit(0);              /* don't dump core */
}

/***********************************************************\
 * Mapping of variable segments
\***********************************************************/

#if !CILK_MANAGE_SHARED_SEGMENT 
/*
 * if linker segments are not used, a portable version
 * of these two functions is provided in cilk.c
 */

/* 
 * map the global variables segment as shared
 */
void Cilk_map_shared_segment(void)
{
     extern int _end_Cilk_shared;	/* linker_defined symbols */
     extern int _begin_Cilk_shared;
     char name[1024];
     int fd;
     void *startmem, *endmem;

     strcpy(name, "/tmp/_Cilk_gvars_XXXXXX");
     mktemp(name);

     startmem = &_begin_Cilk_shared;
     endmem = &_end_Cilk_shared;

     /* write all shared data to the file */
     fd = open(name, O_RDWR | O_CREAT, 0600);
     CILK_CHECK(fd >= 0, ("Open error\n"));
     unlink(name);
     CILK_CHECK(write(fd, startmem, endmem - startmem) == endmem - startmem,
		("error in write\n"));

     /* map shared data to be shared */
     if (startmem != mmap(startmem, endmem - startmem, PROT_READ | PROT_WRITE,
			  MAP_SHARED | MAP_FIXED, fd, 0))
	  Cilk_die("couldn't map shared part to file.\n");
     
     close(fd);
}
#endif

#if !CILK_MANAGE_PRIVATE_SEGMENT 
void Cilk_map_private_segment(void)
{
     /* variables are private by default */
}
#endif

/***********************************************************\
 * Malloc support
\***********************************************************/

/* Get a chunk of memory of size <len> that is mapped shared. */
void *Cilk_map_shared(size_t len)
{
     int fd;
     void *area;
     char name[1024];
     int n;

     strcpy(name, "/tmp/_Cilk_bstore_XXXXXX");
     mktemp(name);

     fd = open(name, O_RDWR | O_CREAT, 0600);
     CILK_CHECK(fd >= 0, ("Open error in Cilk_map_shared\n"));
     unlink(name);

     /* 
      * append something to the end of file, otherwise
      * the system will detonate most spectacularly
      * when trying to access the mmapped memory.
      */
     n = lseek(fd, len, SEEK_SET);
     CILK_CHECK(n == len, ("lseek error in Cilk_map_shared\n"));

     n = write(fd, &len, sizeof(len));
     CILK_CHECK(n == sizeof(len), ("write error in Cilk_map_shared\n"));

     area = mmap(0, len, PROT_READ | PROT_WRITE, 
		 MAP_SHARED, fd, 0);
     if (area == MAP_FAILED) {
	  /* 
	   * can't use Cilk_die, because the mutex in
	   * not initialized yet
	   */
	  fprintf(stderr, "mmap error in Cilk_map_shared -- "
		   "couldn't allocate %d bytes\n", len);
	  exit(1);
     }
     close(fd);
     return area;
}

/***********************************************************\
 * Architecture-specific options
\***********************************************************/

/*
 * Parse any architecture-specific command-line options.  <argc>
 * and <argv> are the original arguments passed on the command line,
 * and <i> is the index of the first architecture-specific option (if any).
 * This function returns the index of the first user option, or <argc> if
 * there aren't any.
 */
int Cilk_arch_specific_options(int i, int argc, char *argv[])
{
     return i;
}

/***********************************************************\
 * Support for mutexen 
\***********************************************************/
#include <asm/system.h>

#define ACQUIRE(lock) xchg((int *)lock, 1)
#define RELEASE(lock) *lock = 0


/* initialize lock */
void Cilk_mutex_init(Cilk_mutex *lock)
{
     *lock = 1;
     RELEASE(lock);
}

/*
 *  Obtain lock <lock>.  Waits until lock is available.
 */
void Cilk_mutex_wait(Cilk_mutex *lock)
{
     if (ACQUIRE(lock) == 0)  /* fast case */
          return;

     Cilk_enter_state(STATE_WAITING_FOR_LOCK);

     /* 
      * This hackery should reduce bus traffic, if the
      * cache is snooped correctly.
      */
     do {
	  while (*lock != 0);
     } while (ACQUIRE(lock) != 0);

     Cilk_exit_state(STATE_WAITING_FOR_LOCK);
}

/*
 *  Try to obtain lock <lock>.  Returns 1 if it succeeds and 0 if it fails.
 *  Does not wait.
 */
int Cilk_mutex_try(Cilk_mutex *lock)
{
     if (*lock != 0)		/*
				 * don't try for lock until we know we
				 * might succeed 
				 */
	  return 0;
     if (ACQUIRE(lock) == 0) {
	  return 1;
     } else
	  return 0;
}

/*
 *  Give up lock <lock>.
 */
void Cilk_mutex_signal(Cilk_mutex *lock)
{
     RELEASE(lock);
}

/***********************************************************\
 * Spawning worker processes
\***********************************************************/

#include <sys/wait.h>
#include <malloc.h>
#include <signal.h>
#include <sys/resource.h>

/*
 * Start forking processes: one process enters this function,
 * <Cilk_active_size> processes should exit this function immediately
 * and return 0.  Each returning process should set <Self> to a unique
 * number between 0 and <Cilk_active_size> - 1.  
 */
void Cilk_create_children(void (*child)(int))
{
     int i, n;
     pid_t pid;
     int stat;
     pid_t *children_pid;


     children_pid = (pid_t *) malloc(USE_SHARED_RO(active_size) 
				     * sizeof(pid_t));
     CILK_CHECK(children_pid,
		("couldn't malloc() children_pid\n"));

     for (i = 0; i < USE_SHARED_RO(active_size); i++) {
	  if ((pid = fork()) < 0) {
	       Cilk_die("can't fork");
	  } else if (pid == 0) {	/* child */
	       child(i);
	       exit(0);
	  } else {		/* parent */
	       children_pid[i] = pid;
	  }
     }

     /* parent now waits for all children to complete */
     for (n = 0; n < USE_SHARED_RO(active_size); n++) {
	  pid = waitpid((pid_t) - 1, &stat, 0);

	  /* find completed child's # */
	  for (i = 0; i < USE_SHARED_RO(active_size); i++)
	       if (pid == children_pid[i])
		    break;
	  CILK_ASSERT(i < USE_SHARED_RO(active_size));

	  if (WIFEXITED(stat)) {
	       if (WEXITSTATUS(stat) == 0) {
		    continue;
	       } else
		    Cilk_dprintf("%d: Child exited with non-zero value %d\n",
				 i, WEXITSTATUS(stat));
	  } else if (WIFSIGNALED(stat)) {
	       const char *core;
	       if (WCOREDUMP(stat)) {
		    core = " (core dumped)";
		    /* prevent others from overwriting core, if possible */
		    Cilk_mutex_try(&USE_SHARED(die_lock));
	       } else
		    core = "";

	       /* 
	        * Print out understandable messages for common signals,
	        * and the signal number for the uncommon ones. 
	        */
	       switch (WTERMSIG(stat)) {
		   case SIGILL:
			Cilk_dprintf("%d: Illegal instruction%s\n", i, core);
			break;
		   case SIGABRT:
			Cilk_dprintf("%d: Abort or Assertion failed%s\n", 
				     i, core);
			break;
		   case SIGFPE:
			Cilk_dprintf("%d: Floating exception%s\n", i, core);
			break;
		   case SIGKILL:
			Cilk_dprintf("%d: Terminated%s\n", i, core);
			break;
		   case SIGBUS:
			Cilk_dprintf("%d: Bus error%s\n", i, core);
			break;
		   case SIGSEGV:
			Cilk_dprintf("%d: Segmentation fault%s\n", i, core);
			break;
		   default:
			Cilk_dprintf(
			  "%d: signal %d (see /usr/include/sys/signal.h)"
					 "%s\n", i, WTERMSIG(stat), core);
			break;
	       }
	  }
	  /* 
	   * something went wrong with child i: abort the rest of the 
	   * children.  Sending a SIGKILL keeps them from dumping core.
	   */
	  for (i = 0; i < USE_SHARED_RO(active_size); i++)
	       kill(children_pid[i], SIGKILL);
     }
}


/**********************************************************
 * Priority control
 **********************************************************/

void Cilk_lower_priority(void)
{
     usleep(10000);
}

void Cilk_raise_priority(void)
{
}

void Cilk_raise_all_priority(void)
{
}

/**********************************************************
 * Initialization of the module
 **********************************************************/

void Cilk_arch_specific_init(void)
{
     Cilk_mutex_init(&USE_SHARED(dprintf_lock));
     Cilk_mutex_init(&USE_SHARED(die_lock));
}

void Cilk_arch_specific_per_process_init(void)
{
}

