/*
 * $Id: arch-solaris-threads.c,v 1.2 1998/06/01 14:32:47 fengmd Exp $
 */
/*
 * $Log: arch-solaris-threads.c,v $
 * Revision 1.2  1998/06/01 14:32:47  fengmd
 * Merge changes from the main chunk.
 *
 * Revision 1.1.1.1  1998/03/19 04:43:35  astark
 * Initial creation of umbrella/all-sets implementation
 *
 * Revision 1.1.1.1.2.2  1997/06/16 16:00:23  fengmd
 * Remerge from the base for Version 5.0.1.
 *
 * Revision 1.4  1997/06/09 19:03:05  athena
 * Changed RTS so that Cilk won't dump core by default
 *
 * Revision 1.3  1997/03/02 21:06:25  randall
 * Made sure Cilk_die doesn't return.
 *
 * Revision 1.2  1997/03/01 17:42:31  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <stdarg.h>
#include <stdio.h>
#include <cilk.h>
#include <cilk-internal.h>
#include <cilk-lib.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/priocntl.h>
#include <sys/tspriocntl.h>
#include <sys/procfs.h>
#include <pthread.h>

DECLARE_SHARED_RO_VARIABLE(static int, lock_type, );
#define MEMORY_LOCKS 0
#define OS_LOCKS 1

/***********************************************************\
 * Debugging & tracing
\***********************************************************/

DECLARE_SHARED_VARIABLE(static Cilk_mutex, dprintf_lock,);

void Cilk_dprintf(const char *fmt,...)
{
     va_list l;

     /* ensure nice interleaving of debugging messages */
     Cilk_mutex_wait(&USE_SHARED(dprintf_lock));
     if (Self >= 0)
	  fprintf(stderr, "%d: ", Self);
     else
	  fprintf(stderr, "global: ");
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     fflush(stderr);
     va_end(l);
     Cilk_mutex_signal(&USE_SHARED(dprintf_lock));
}

DECLARE_SHARED_VARIABLE(static Cilk_mutex, die_lock,);

void Cilk_die(const char *fmt,...)
{
     va_list l;

     Cilk_mutex_wait(&USE_SHARED(dprintf_lock));
     if (Self >= 0)
	  fprintf(stderr, "%d: fatal error: ", Self);
     else
	  fprintf(stderr, "global: fatal error: ");
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     va_end(l);
     Cilk_mutex_signal(&USE_SHARED(dprintf_lock));

     /* ensure only the first fatal error produces death of the program */
     if (USE_SHARED_RO(dump_core) && Cilk_mutex_try(&USE_SHARED(die_lock)))
	  abort();		/* dump core */
     else
          exit(0);              /* don't dump core */
}

/***********************************************************\
 * Mapping of variable segments
\***********************************************************/

#if !CILK_MANAGE_SHARED_SEGMENT
/*
 * if linker segments are not used, a portable version
 * of these two functions is provided in cilk.c
 */

/* 
 * map the global variables segment as shared
 */
void Cilk_map_shared_segment(void)
{
     /* nothing needs to be done */
}

#endif

#if !CILK_MANAGE_PRIVATE_SEGMENT
void Cilk_map_private_segment(void)
{
     /* variables are private by default */
}

#endif

/***********************************************************\
 * Malloc support
\***********************************************************/
/*
 * we could use malloc() instead of mmap(), since under
 * Solaris threads malloc() gives back shared memory anyway.
 * However, in case of core dump, the whole heap is dumped
 * (64Mb?), which is too much.
 */
void *Cilk_map_shared(size_t len)
{
     int fd;
     void *area;

     fd = open("/dev/zero", O_RDWR);

     CILK_CHECK(fd >= 0, ("Open error in Cilk_map_shared\n"));

     area = mmap(0, len, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
     if (area == MAP_FAILED) {
	  /* 
	   * can't use Cilk_die, because the mutex in
	   * not initialized yet
	   */
	  fprintf(stderr, "mmap error in Cilk_map_shared -- "
		   "couldn't allocate %d bytes\n", len);
	  exit(1);
     }
     close(fd);
     return area;
}

/***********************************************************\
 * Architecture-specific options
\***********************************************************/

/*
 * Parse any architecture-specific command-line options.  <argc>
 * and <argv> are the original arguments passed on the command line,
 * and <i> is the index of the first architecture-specific option (if any).
 * This function returns the index of the first user option, or <argc> if
 * there aren't any.
 */
int Cilk_arch_specific_options(int i, int argc, char *argv[])
{
     USE_SHARED_RO(lock_type) = MEMORY_LOCKS;

     for (; i < argc; ++i) {
	  if (*argv[i] != '-')
	       break;

	  if (strcmp(argv[i] + 1, "-") == 0) {
	       ++i;
	       break;
	  } else if (strcmp(argv[i] + 1, "os-locks") == 0) {
	       USE_SHARED_RO(lock_type) = OS_LOCKS;
	  } else if (strcmp(argv[i] + 1, "memory-locks") == 0) {
	       USE_SHARED_RO(lock_type) = MEMORY_LOCKS;
	  } else
	       break;
     }

     return i;
}

/***********************************************************\
 * Support for mutexen 
\***********************************************************/

#if SPARC_V9
#define ACQUIRE(lock) Cilk_cas(lock, 0, 1)
#define RELEASE(lock) *lock = 0
#endif

#if SPARC_V8
#define ACQUIRE(lock) Cilk_swap(lock, 1)
#define RELEASE(lock) Cilk_swap(lock, 0)
#endif

/*******************************
 * Memory locks
 *******************************/
/* initialize lock */
static inline void Cilk_mutex_init_memory(Cilk_mutex *lock)
{
     lock->memory = 1;
     RELEASE(&lock->memory);
}

/*
 *  Obtain lock <lock>.  Waits until lock is available.
 */
static inline void Cilk_mutex_wait_memory(Cilk_mutex *lock)
{
     if (ACQUIRE(&lock->memory) == 0)  /* fast case */
          return;

     Cilk_enter_state(STATE_WAITING_FOR_LOCK);

     /* 
      * This hackery should reduce bus traffic, if the
      * cache is snooped correctly.
      */
     do {
	  while (lock->memory != 0);
     } while (ACQUIRE(&lock->memory) != 0);

     Cilk_exit_state(STATE_WAITING_FOR_LOCK);
}

/*
 *  Give up lock <lock>.
 */
static inline void Cilk_mutex_signal_memory(Cilk_mutex *lock)
{
     RELEASE(&lock->memory);
}

/*
 *  Try to obtain lock <lock>.  Returns 1 if it succeeds and 0 if it fails.
 *  Does not wait.
 */
static inline int Cilk_mutex_try_memory(Cilk_mutex *lock)
{
     if (ACQUIRE(&lock->memory) == 0) {
	  return 1;
     } else
	  return 0;
}

/*******************************
 * Operating System locks
 *******************************/
/* initialize lock */
static inline void Cilk_mutex_init_os(Cilk_mutex *lock)
{
     mutex_init(&lock->os, USYNC_PROCESS, NULL);
}

/*
 *  Obtain lock <lock>.  Waits until lock is available.
 */
static inline void Cilk_mutex_wait_os(Cilk_mutex *lock)
{
     Cilk_enter_state(STATE_WAITING_FOR_LOCK);

     mutex_lock(&lock->os);

     Cilk_exit_state(STATE_WAITING_FOR_LOCK);
}

/*
 *  Give up lock <lock>.
 */
static inline void Cilk_mutex_signal_os(Cilk_mutex *lock)
{
     mutex_unlock(&lock->os);
}

/*
 *  Try to obtain lock <lock>.  Returns 1 if it succeeds and 0 if it fails.
 *  Does not wait.
 */
static inline int Cilk_mutex_try_os(Cilk_mutex *lock)
{
     if (mutex_trylock(&lock->os) == 0) {
	  return 1;
     } else {
	  return 0;
     }
}

/*********************
 * Lock glue 
 *********************/
#define MAKE_LOCKFUN(name) \
void name(Cilk_mutex *lock)                                   \
{                                                             \
     if (USE_SHARED_RO(lock_type) == MEMORY_LOCKS) {	      \
	  name##_memory(lock);				      \
     } else {						      \
	  name##_os(lock);				      \
     }							      \
}							      \
							       
MAKE_LOCKFUN(Cilk_mutex_init)
MAKE_LOCKFUN(Cilk_mutex_wait)
MAKE_LOCKFUN(Cilk_mutex_signal)

int Cilk_mutex_try(Cilk_mutex *lock)
{
     if (USE_SHARED_RO(lock_type) == MEMORY_LOCKS) {
	  return Cilk_mutex_try_memory(lock);
     } else {
	  return Cilk_mutex_try_os(lock);
     }
}

/***********************************************************\
 * Spawning worker processes
\***********************************************************/

#include <sys/wait.h>
#include <malloc.h>
#include <signal.h>
#include <sys/resource.h>

DECLARE_SHARED_VARIABLE(static void *, shared_tid, );

/*
 * Join all the children
 */
static void join_children(void) {
    int i;
    pthread_t *tid = USE_SHARED(shared_tid);

    for ( i = 0; i < USE_SHARED_RO(active_size); i++)
        pthread_join(tid[i], NULL);
}

/*
 * Start forking processes: one process enters this function,
 * <Cilk_active_size> processes should exit this function immediately
 * and return 0.  Each returning process should set <Self> to a unique
 * number between 0 and <Cilk_active_size> - 1.
 */
void Cilk_create_children(void (*child) (int))
{
     int i;
     int res;
     pthread_attr_t attr;
     pthread_t *tid;

     tid = malloc(USE_SHARED_RO(active_size) * sizeof(pthread_t));
     CILK_CHECK(tid, ("couldn't children() tid\n"));
     USE_SHARED(shared_tid) = tid;

     pthread_attr_init(&attr); /* initialize attr with default attributes */
     pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);  /* system-wide contention */

     for (i = 0; i < USE_SHARED_RO(active_size); i++) {
          res = pthread_create(&tid[i], 
			       &attr,
			       (void * (*) (void *)) child, 
			       (void *) i);
	  if (res) {
	       Cilk_die("Can't create threads\n");
	  }
     }

#ifndef CILK_LIBRARY_VERSION
     join_children();
#endif
}

/***************************************************
 * Timer control
 ***************************************************/
static int openslashproc(int pid)
{
     char buf[64];
     int fd;

     sprintf(buf, "/proc/%d", pid);
     if ((fd = open(buf, 2)) < 0)
	  perror("open /proc");

     return fd;
}

static void start_microaccounting(void)
{
     int fd, flags;

     fd = openslashproc(getpid());

     flags = PR_MSACCT;
     if (ioctl(fd, PIOCSET, &flags, sizeof(flags)) < 0) {
	  perror("ioctl set microstate acct");
     }
}

/**********************************************************
 * Priority control
 **********************************************************/
#define PRIORITY_DROP 20

DECLARE_PRIVATE_VARIABLE(static int, current_priority,);

void Cilk_lower_priority(void)
{
     struct pcparms params;
     struct tsparms *tsparams = (struct tsparms *) &params.pc_clparms;

     if (USE_PRIVATE(current_priority) == 0) {
	  params.pc_cid = PC_CLNULL;
	  priocntl(P_LWPID, _lwp_self(), PC_GETPARMS, (caddr_t) & params);
	  USE_PRIVATE(current_priority) = 1;
	  tsparams->ts_upri = tsparams->ts_uprilim - PRIORITY_DROP;
	  priocntl(P_LWPID, _lwp_self(), PC_SETPARMS, (caddr_t) & params);
     }
#ifndef ARCH_NO_YIELDS
     else {
       if ((USE_PRIVATE(current_priority)++) ==
	   2 * USE_SHARED_RO(active_size)) {
	 USE_PRIVATE(current_priority) = 1;
	 Cilk_yield();
       }
     }
#endif /* ifndef ARCH_NO_YIELDS */
}

void Cilk_raise_priority(void)
{
     struct pcparms params;
     struct tsparms *tsparams = (struct tsparms *) &params.pc_clparms;

     if (USE_PRIVATE(current_priority) != 0) {
	  params.pc_cid = PC_CLNULL;
	  priocntl(P_LWPID, _lwp_self(), PC_GETPARMS, (caddr_t) & params);
	  USE_PRIVATE(current_priority) = 0;
	  tsparams->ts_upri = tsparams->ts_uprilim;
	  priocntl(P_LWPID, _lwp_self(), PC_SETPARMS, (caddr_t) & params);
     }
}

/* Bump up all worker thread priorities.  (The threads may still think
 * they have low priority.  One Cilk_raise_priority followed by a
 * Cilk_lower_priority will fix that.)  There might be race conditions
 * with the workers: the worker may immediately lower its priority
 * after the call the Cilk_raise_all_priority.
 */
void Cilk_raise_all_priority(void)
{
     int i;
     struct pcparms params;
     struct tsparms *tsparams = (struct tsparms *) &params.pc_clparms;
     pthread_t *tid = USE_SHARED(shared_tid);

     for (i = 0; i < USE_SHARED_RO(active_size); i++) {
          params.pc_cid = PC_CLNULL;
	  priocntl(P_LWPID, tid[i], PC_GETPARMS, (caddr_t) & params);
	  tsparams->ts_upri = tsparams->ts_uprilim;
	  priocntl(P_LWPID, tid[i], PC_SETPARMS, (caddr_t) & params);
     }
}

#ifdef CILK_LIBRARY_VERSION
/**********************************************************
 * Support for library version
 **********************************************************/

DECLARE_SHARED_VARIABLE(static mutex_t, library_state_mutex, );
DECLARE_SHARED_VARIABLE(static cond_t, library_state_condvar, );
DECLARE_SHARED_VARIABLE(static int, library_state, );
DECLARE_SHARED_VARIABLE(static volatile int, library_count, );
#define CILK_LIBRARY_IDLE 0
#define CILK_LIBRARY_RUNNING 1
#define CILK_LIBRARY_EXITING 2

void Cilk_library_init(void)
{
  /* initialize library condition variables */
  USE_SHARED(library_state) = CILK_LIBRARY_IDLE;
  USE_SHARED(library_count) = 0;
  mutex_init(&USE_SHARED(library_state_mutex), USYNC_THREAD, 0);
  cond_init(&USE_SHARED(library_state_condvar), USYNC_THREAD, 0);
}

void Cilk_library_exit(void)
{
  Cilk_raise_all_priority();  /* one last time */

  USE_SHARED(exit) = 1;
  
  Cilk_fence();

  mutex_lock(&USE_SHARED(library_state_mutex));
  USE_SHARED(library_state) = CILK_LIBRARY_EXITING;
  cond_broadcast(&USE_SHARED(library_state_condvar));
  mutex_unlock(&USE_SHARED(library_state_mutex));

  join_children();
}

void Cilk_library_ready_to_run(void)
{
  /* wait to make sure all worker threads are asleep after last Cilk_run */
  while(USE_SHARED(library_count) > 0) {
    Cilk_raise_all_priority();
    Cilk_yield();
  }
}

void Cilk_library_run(void)
{
  /* wait to make sure all worker threads are asleep after last Cilk_run */
  while(USE_SHARED(library_count) > 0)
    Cilk_yield();

  Cilk_fence();

  CILK_ASSERT(USE_SHARED(library_state) == CILK_LIBRARY_IDLE);

  USE_SHARED(done) = 0;
  
  Cilk_fence();

  /*
   * Wake up cilk workers
   */
  mutex_lock(&USE_SHARED(library_state_mutex));
  CILK_ASSERT(USE_SHARED(library_state) == CILK_LIBRARY_IDLE);
  USE_SHARED(library_state) = CILK_LIBRARY_RUNNING;
  USE_SHARED(library_count) = USE_SHARED_RO(active_size);
  cond_broadcast(&USE_SHARED(library_state_condvar));
  mutex_unlock(&USE_SHARED(library_state_mutex));
}

void Cilk_library_main_sleep(void)
{
  /*
   * Sleep this thread until the work is done
   */
  mutex_lock(&USE_SHARED(library_state_mutex));
  while(USE_SHARED(library_state) != CILK_LIBRARY_IDLE)
    {
      cond_wait(&USE_SHARED(library_state_condvar),
		&USE_SHARED(library_state_mutex));
    }
  mutex_unlock(&USE_SHARED(library_state_mutex));
}

void Cilk_library_main_wakeup(void)
{
     /* wait to make sure all worker threads wake up */
     while(USE_SHARED(library_count) > 0)
       Cilk_yield();
     
     Cilk_fence();

     /* signal scheduling loops to exit */
     USE_SHARED(done) = 1;
     
     Cilk_fence();

     /* signal parent thread to resume */
     mutex_lock(&USE_SHARED(library_state_mutex));
     USE_SHARED(library_state) = CILK_LIBRARY_IDLE;
     USE_SHARED(library_count) = USE_SHARED_RO(active_size);
     cond_broadcast(&USE_SHARED(library_state_condvar));
     mutex_unlock(&USE_SHARED(library_state_mutex));

     /* raise all worker priorities */
     Cilk_raise_all_priority();
}

/*
 * Called from the scheduling loop
 */
void Cilk_library_wait_for_wakeup(void)
{
     mutex_lock(&USE_SHARED(library_state_mutex));
     while(USE_SHARED(library_state) == CILK_LIBRARY_IDLE)
       {
	 cond_wait(&USE_SHARED(library_state_condvar),
		   &USE_SHARED(library_state_mutex));
       }
     /* record that we saw the go flag */
     USE_SHARED(library_count)--;
     mutex_unlock(&USE_SHARED(library_state_mutex));

     Cilk_fence();
}

void Cilk_library_go_to_sleep(void)
{
     /* record that we saw the done flag */
     mutex_lock(&USE_SHARED(library_state_mutex));
     USE_SHARED(library_count)--;
     mutex_unlock(&USE_SHARED(library_state_mutex));
     
     Cilk_fence();
     /* actual sleeping does not occur until next wait_for_wakeup call */
}
#endif /* ifdef CILK_LIBRARY_VERSION */

/**********************************************************
 * Initialization of the module
 **********************************************************/

void Cilk_arch_specific_init(void)
{
     Cilk_mutex_init(&USE_SHARED(dprintf_lock));
     Cilk_mutex_init(&USE_SHARED(die_lock));
     start_microaccounting();
}

void Cilk_arch_specific_per_process_init(void)
{
     USE_PRIVATE(current_priority) = -1;
     start_microaccounting();
}
