/*
 * $Id: arch-dunix-threads.c,v 1.6 1998/05/21 22:34:48 lisiecki Exp $
 */
/*
 * $Log: arch-dunix-threads.c,v $
 * Revision 1.6  1998/05/21 22:34:48  lisiecki
 * library version missing function
 *
 * Revision 1.5  1998/05/21 21:40:27  lisiecki
 * Library version support
 *
 * Revision 1.4  1997/11/18 19:12:18  randall
 * Fix to reduce memory bandwidth consumed by stealers.
 *
 * Revision 1.3  1997/09/23 19:29:27  randall
 * Bug fix.
 *
 * Revision 1.2  1997/09/22 20:27:17  randall
 * Got process version of Cilk working.  Shared globals still aren't working!!!
 *
 * Revision 1.1  1997/09/17 19:07:29  randall
 * Added digital unix thread version.
 *
 * Revision 1.1.1.1  1997/02/14 23:07:12  strumpen
 * Create Cilk Repository
 *
 * Revision 1.1.1.1  1997/02/14 22:04:29  strumpen
 * Create cilk repository
 *
 * Revision 1.1  1997/01/16 14:23:03  athena
 * Added support for linux
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <stdarg.h>
#include <stdio.h>
#include <cilk.h>
#include <cilk-internal.h>
#include <cilk-lib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>

DECLARE_SHARED_RO_VARIABLE(static int, lock_type, );
#define MEMORY_LOCKS 0
#define OS_LOCKS 1

/***********************************************************\
 * Debugging & tracing
\***********************************************************/

DECLARE_SHARED_VARIABLE(static Cilk_mutex, dprintf_lock, );

void Cilk_dprintf(const char *fmt,...)
{
     va_list l;

     /* ensure nice interleaving of debugging messages */
     Cilk_mutex_wait(&USE_SHARED(dprintf_lock));
     if (Self >= 0)
	  fprintf(stderr, "%d: ", Self);
     else
	  fprintf(stderr, "global: ");
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     fflush(stderr);
     va_end(l);
     Cilk_mutex_signal(&USE_SHARED(dprintf_lock));
}

DECLARE_SHARED_VARIABLE(static Cilk_mutex, die_lock,);

void Cilk_die(const char *fmt,...)
{
     va_list l;

     Cilk_mutex_wait(&USE_SHARED(dprintf_lock));
     if (Self >= 0)
	  fprintf(stderr, "%d: fatal error: ", Self);
     else
	  fprintf(stderr, "global: fatal error: ");
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     va_end(l);
     Cilk_mutex_signal(&USE_SHARED(dprintf_lock));

     /* ensure only the first fatal error produces death of the program */
     if (Cilk_mutex_try(&USE_SHARED(die_lock)))
	  abort();		/* dump core */
}

/***********************************************************\
 * Mapping of variable segments
\***********************************************************/

#if !CILK_MANAGE_SHARED_SEGMENT
/*
 * if linker segments are not used, a portable version
 * of these two functions is provided in cilk.c
 */

/* 
 * map the global variables segment as shared
 */
void Cilk_map_shared_segment(void)
{
     /* nothing needs to be done */
}
#endif

#if !CILK_MANAGE_PRIVATE_SEGMENT 
void Cilk_map_private_segment(void)
{
     /* variables are private by default */
}
#endif

/***********************************************************\
 * Malloc support
\***********************************************************/
/*
 * we could use malloc() instead of mmap(), since under
 * the threaded version, malloc() gives back shared memory anyway.
 * However, in case of core dump, the whole heap is dumped
 * (64Mb?), which is too much.
 */
void *Cilk_map_shared(size_t len)
{
     int fd;
     void *area;

     fd = open("/dev/zero", O_RDWR);

     CILK_CHECK(fd >= 0, ("Open error in Cilk_map_shared\n"));

     area = mmap(0, len, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
     if (((long) area) == MAP_FAILED) {
	  /* 
	   * can't use Cilk_die, because the mutex in
	   * not initialized yet
	   */
	  fprintf(stderr, "mmap error in Cilk_map_shared -- "
		   "couldn't allocate %ld bytes\n", len);
	  exit(1);
     }
     close(fd);
     return area;
}

/***********************************************************\
 * Architecture-specific options
\***********************************************************/

/*
 * Parse any architecture-specific command-line options.  <argc>
 * and <argv> are the original arguments passed on the command line,
 * and <i> is the index of the first architecture-specific option (if any).
 * This function returns the index of the first user option, or <argc> if
 * there aren't any.
 */
int Cilk_arch_specific_options(int i, int argc, char *argv[])
{
     USE_SHARED_RO(lock_type) = MEMORY_LOCKS;

     for (; i < argc; ++i) {
	  if (*argv[i] != '-')
	       break;

	  if (strcmp(argv[i] + 1, "-") == 0) {
	       ++i;
	       break;
	  } else if (strcmp(argv[i] + 1, "os-locks") == 0) {
	       USE_SHARED_RO(lock_type) = OS_LOCKS;
	  } else if (strcmp(argv[i] + 1, "memory-locks") == 0) {
	       USE_SHARED_RO(lock_type) = MEMORY_LOCKS;
	  } else
	       break;
     }

     return i;
}

/***********************************************************\
 * Timer support
\***********************************************************/

DECLARE_PRIVATE_VARIABLE(static Cilk_time, Cilk_last_clock_low32, );
DECLARE_PRIVATE_VARIABLE(static Cilk_time, Cilk_last_clock_high32, );

/* Note: this returns a virtual time modulo 2^32.  Since we are
 * counting at a rate of 500MHz, the counter overflows in only 8
 * seconds!  This is not good enough for real life, but for now it is
 * passable.  Critical path measurements will work if all thread
 * lengths are small, because time is still accumulated in a 64-bit
 * value.
 */
Cilk_time Cilk_get_time(void)
{
  register Cilk_time clock_low32;
  register Cilk_time tmp;
  
  /* Read low 32 bits of clock.
   * From page 4-142 in the Alpha Architecture Handbook
   * http://ftp.digital.com/pub/Digital/info/semiconductor/literature/alphahb2.pdf
   */
  asm ("rpcc %0\n"
       "sll %0, 32, %1\n"
       "addq %0, %1, %0\n"
       "srl %0, 32, %0" : "=r" (clock_low32), "=r" (tmp) : );
  
  /* if we have wrapped around, add to high 32 bits of clock */
  if (clock_low32 < USE_PRIVATE(Cilk_last_clock_low32))
    {
      /* overflow occurred - fix up */
      USE_PRIVATE(Cilk_last_clock_high32) += (1LL<<32);
    }
  
  /* save most up-to-date low 32 bits of clock */
  USE_PRIVATE(Cilk_last_clock_low32) = clock_low32;
  
  /* return clock value */
  return USE_PRIVATE(Cilk_last_clock_high32) + clock_low32;
}

/***********************************************************\
 * Support for mutexen 
\***********************************************************/

/*******************************
 * Memory locks
 *   still need to check that this implementation is correct -cfj
 *******************************/
/* initialize lock */
void Cilk_mutex_init_memory(Cilk_mutex *lock)
{
     lock->memory = 0;
     CILK_W_SYNC;
}

/*
 *  Obtain lock <lock>.  Waits until lock is available.
 */
void Cilk_mutex_wait_memory(Cilk_mutex *lock)
{
     register int tmp;
     Cilk_enter_state(STATE_WAITING_FOR_LOCK);
     /*
  while(1)
    asm volatile ("loop: ldl   %0, (%1)\n"
		  "blbs  %0, loop\n"
		  "ldl_l %0, (%1)\n"
		  "blbs  %0, loop\n"
		  "or    %0, 1, %0\n"
		  "stl_c %0, (%1)\n"
		  "beq   %0, loop\n" : "=r" (tmp) : "r" (lock));
  */
     do {
          do {
               asm volatile ("ldl_l %0, (%1)":"=r" (tmp):"r"(&lock->memory));
          } while (tmp != 0);

          tmp = 1;
          asm volatile ("stl_c %0, (%1)":"=r" (tmp):"r"(&lock->memory), "0"(tmp));
     } while (tmp == 0);
     
     CILK_SYNC;
     Cilk_exit_state(STATE_WAITING_FOR_LOCK);
}

/*
 *  Try to obtain lock <lock>.  Returns 1 if it succeeds and 0 if it fails.
 *  Does not wait.
 */
int Cilk_mutex_try_memory(Cilk_mutex *lock)
{
     int tmp;

     asm volatile ("ldl_l %0, (%1)":"=r" (tmp):"r"(&lock->memory));
     if (tmp != 0)
          return 0;

     tmp = 1;
     asm volatile ("stl_c %0, (%1)":"=r" (tmp):"r"(&lock->memory), "0"(tmp));
     CILK_SYNC;
     return tmp;
}

/*
 *  Give up lock <lock>.
 */
void Cilk_mutex_signal_memory(Cilk_mutex *lock)
{
     CILK_SYNC;
     lock->memory = 0;
}


/*******************************
 * Operating System locks
 *******************************/
/*
 * Initialize lock
 */
static inline void Cilk_mutex_init_os(Cilk_mutex *lock)
{
    pthread_mutex_init(&lock->os, NULL);
}

/*
 *  Obtain lock <lock>.  Waits until lock is available.
 */
static inline void Cilk_mutex_wait_os(Cilk_mutex *lock)
{
     Cilk_enter_state(STATE_WAITING_FOR_LOCK);

     pthread_mutex_lock(&lock->os);

     Cilk_exit_state(STATE_WAITING_FOR_LOCK);
}

/*
 *  Give up lock <lock>.
 */
static inline void Cilk_mutex_signal_os(Cilk_mutex *lock)
{
     pthread_mutex_unlock(&lock->os);
}

/*
 *  Try to obtain lock <lock>.  Returns 1 if it succeeds and 0 if it fails.
 *  Does not wait.
 */
static inline int Cilk_mutex_try_os(Cilk_mutex *lock)
{
     if (pthread_mutex_trylock(&lock->os) == 0) {
	  return 1;
     } else {
	  return 0;
     }
}

/*********************
 * Lock glue 
 *********************/

#define MAKE_LOCKFUN(name) \
void name(Cilk_mutex *lock)                                   \
{                                                             \
     if (USE_SHARED_RO(lock_type) == MEMORY_LOCKS) {	      \
	  name##_memory(lock);				      \
     } else {						      \
	  name##_os(lock);				      \
     }							      \
}							      \

MAKE_LOCKFUN(Cilk_mutex_init)
MAKE_LOCKFUN(Cilk_mutex_wait)
MAKE_LOCKFUN(Cilk_mutex_signal)

int Cilk_mutex_try(Cilk_mutex *lock)
{
     if (USE_SHARED_RO(lock_type) == MEMORY_LOCKS) {
	  return Cilk_mutex_try_memory(lock);
     } else {
	  return Cilk_mutex_try_os(lock);
     }
}


/***********************************************************\
 * Spawning worker processes
\***********************************************************/

/*
 * Start forking processes: one process enters this function,
 * <active_size> processes should call <child> immediately, with
 * an argument between 0 and <active_size>-1.  This function should
 * return only when all of those child calls have completed.
 */
void Cilk_create_children(void (*child)(int))
{
  long i;
  int res;
  pthread_t *tid;
  
  tid = malloc(USE_SHARED_RO(active_size) * sizeof(pthread_t));
  
  for (i = 0; i < USE_SHARED_RO(active_size); i++)
    {
      res = pthread_create(tid + i, 
			   NULL,
			   (void * (*) (void *)) child, 
			   (void *) i);
      if (res)
	Cilk_die("Can't create threads\n");
    }

#ifdef CILK_LIBRARY_VERSION
  return;
#endif

  for (i = 0; i < USE_SHARED_RO(active_size); i++)
    {
      res = pthread_join(tid[i], NULL);
      CILK_CHECK(res == 0,
		 ("error in join: %d returned!!\n", res));
    }
}

/**********************************************************
 * Priority control
 **********************************************************/

#define NO_PRIORITY 0
#define HIGH_PRIORITY 1
#define LOW_PRIORITY 2
DECLARE_PRIVATE_VARIABLE(static int, current_priority,);

void Cilk_lower_priority(void)
{
  if (USE_PRIVATE(current_priority) < LOW_PRIORITY)
    {
      struct sched_param param;
      param.sched_priority = PRI_OTHER_MIN;
      pthread_setschedparam(pthread_self(), SCHED_OTHER, &param);
      USE_PRIVATE(current_priority) = LOW_PRIORITY;
    }
#ifndef ARCH_NO_YIELDS
  else {
    if ((USE_PRIVATE(current_priority)++) ==
	LOW_PRIORITY + 2 * USE_SHARED_RO(active_size)) {
      USE_PRIVATE(current_priority) = LOW_PRIORITY;
      Cilk_yield();
    }
  }
#endif /* ifndef ARCH_NO_YIELDS */
}

void Cilk_raise_priority(void)
{
  if (USE_PRIVATE(current_priority) != HIGH_PRIORITY)
    {
      struct sched_param param;
      param.sched_priority = PRI_OTHER_MAX;
      pthread_setschedparam(pthread_self(), SCHED_OTHER, &param);
      USE_PRIVATE(current_priority) = HIGH_PRIORITY;
    }
}

void Cilk_raise_all_priority(void)
{
  /* Does not seem to be necessary */
}

#ifdef CILK_LIBRARY_VERSION
/**********************************************************
 * Support for library version
 **********************************************************/

DECLARE_SHARED_VARIABLE(static pthread_mutex_t, library_state_mutex, );
DECLARE_SHARED_VARIABLE(static pthread_cond_t, library_state_condvar, );
DECLARE_SHARED_VARIABLE(static int, library_state, );
DECLARE_SHARED_VARIABLE(static volatile int, library_count, );
#define CILK_LIBRARY_IDLE 0
#define CILK_LIBRARY_RUNNING 1
#define CILK_LIBRARY_EXITING 2

void Cilk_library_init(void)
{
  /* initialize library condition variables */
  USE_SHARED(library_state) = CILK_LIBRARY_IDLE;
  USE_SHARED(library_count) = 0;
  pthread_mutex_init(&USE_SHARED(library_state_mutex), NULL);
  pthread_cond_init(&USE_SHARED(library_state_condvar), NULL);
}

void Cilk_library_exit(void)
{
  USE_SHARED(exit) = 1;
  
  pthread_mutex_lock(&USE_SHARED(library_state_mutex));
  USE_SHARED(library_state) = CILK_LIBRARY_EXITING;
  pthread_cond_broadcast(&USE_SHARED(library_state_condvar));
  pthread_mutex_unlock(&USE_SHARED(library_state_mutex));
}

void Cilk_library_ready_to_run(void)
{
  /* wait to make sure all worker threads are asleep after last Cilk_run */
  while(USE_SHARED(library_count) > 0) {
    Cilk_raise_all_priority();
    Cilk_yield();
  }
}

void Cilk_library_run(void)
{
  /* wait to make sure all worker threads are asleep after last Cilk_run */
  while(USE_SHARED(library_count) > 0)
    Cilk_yield();

  Cilk_fence();

  CILK_ASSERT(USE_SHARED(library_state) == CILK_LIBRARY_IDLE);

  USE_SHARED(done) = 0;  

  Cilk_fence();

  /*
   * Wake up cilk workers
   */
  pthread_mutex_lock(&USE_SHARED(library_state_mutex));
  CILK_ASSERT(USE_SHARED(library_state) == CILK_LIBRARY_IDLE);
  USE_SHARED(library_state) = CILK_LIBRARY_RUNNING;
  USE_SHARED(library_count) = USE_SHARED_RO(active_size);
  pthread_cond_broadcast(&USE_SHARED(library_state_condvar));
  pthread_mutex_unlock(&USE_SHARED(library_state_mutex));
}

void Cilk_library_main_sleep(void)
{
  /*
   * Sleep this thread until the work is done
   */
  pthread_mutex_lock(&USE_SHARED(library_state_mutex));
  while(USE_SHARED(library_state) != CILK_LIBRARY_IDLE)
    {
      pthread_cond_wait(&USE_SHARED(library_state_condvar),
			&USE_SHARED(library_state_mutex));
    }
  pthread_mutex_unlock(&USE_SHARED(library_state_mutex));
}

void Cilk_library_main_wakeup(void)
{
     /* wait to make sure all worker threads wake up */
     while(USE_SHARED(library_count) > 0);
     
     /* signal scheduling loops to exit */
     USE_SHARED(done) = 1;
     
     /* signal parent thread to resume */
     pthread_mutex_lock(&USE_SHARED(library_state_mutex));
     USE_SHARED(library_state) = CILK_LIBRARY_IDLE;
     USE_SHARED(library_count) = USE_SHARED_RO(active_size);
     pthread_cond_broadcast(&USE_SHARED(library_state_condvar));
     pthread_mutex_unlock(&USE_SHARED(library_state_mutex));
}

/*
 * Called from the scheduling loop
 */
void Cilk_library_wait_for_wakeup(void)
{
     pthread_mutex_lock(&USE_SHARED(library_state_mutex));
     while(USE_SHARED(library_state) != CILK_LIBRARY_RUNNING)
       {
	 pthread_cond_wait(&USE_SHARED(library_state_condvar),
			   &USE_SHARED(library_state_mutex));
       }
     /* record that we saw the go flag */
     USE_SHARED(library_count)--;
     pthread_mutex_unlock(&USE_SHARED(library_state_mutex));
}

void Cilk_library_go_to_sleep(void)
{
     /* record that we saw the done flag */
     pthread_mutex_lock(&USE_SHARED(library_state_mutex));
     USE_SHARED(library_count)--;
     pthread_mutex_unlock(&USE_SHARED(library_state_mutex));
     
     /* actual sleeping does not occur until next wait_for_wakeup call */
}
#endif /* ifdef CILK_LIBRARY_VERSION */

/**********************************************************
 * Initialization of the module
 **********************************************************/

void Cilk_arch_specific_init(void)
{
     Cilk_mutex_init(&USE_SHARED(dprintf_lock));
     Cilk_mutex_init(&USE_SHARED(die_lock));
}

void Cilk_arch_specific_per_process_init(void)
{
     USE_PRIVATE(Cilk_last_clock_low32) = 0;
     USE_PRIVATE(Cilk_last_clock_high32) = 0;
     USE_PRIVATE(current_priority) = NO_PRIORITY;
}

