/*
 * $Id: arch-dunix.c,v 1.2 1998/05/21 21:40:28 lisiecki Exp $
 */
/*
 * $Log: arch-dunix.c,v $
 * Revision 1.2  1998/05/21 21:40:28  lisiecki
 * Library version support
 *
 * Revision 1.1  1997/09/22 20:27:20  randall
 * Got process version of Cilk working.  Shared globals still aren't working!!!
 *
 * Revision 1.1.1.1  1997/02/14 23:07:12  strumpen
 * Create Cilk Repository
 *
 * Revision 1.1.1.1  1997/02/14 22:04:29  strumpen
 * Create cilk repository
 *
 * Revision 1.1  1997/01/16 14:23:03  athena
 * Added support for linux
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <stdarg.h>
#include <stdio.h>
#include <cilk.h>
#include <cilk-internal.h>
#include <cilk-lib.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <unistd.h>

//#include <sys/ipc.h>
//#include <sys/shm.h>

DECLARE_SHARED_RO_VARIABLE(static int, lock_type, );
#define MEMORY_LOCKS 0
#define OS_LOCKS 1

/***********************************************************\
 * Debugging & tracing
\***********************************************************/

DECLARE_SHARED_VARIABLE(static Cilk_mutex, dprintf_lock, );

void Cilk_dprintf(const char *fmt,...)
{
     va_list l;

     /* ensure nice interleaving of debugging messages */
     Cilk_mutex_wait(&USE_SHARED(dprintf_lock));
     if (Self >= 0)
	  fprintf(stderr, "%d: ", Self);
     else
	  fprintf(stderr, "global: ");
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     fflush(stderr);
     va_end(l);
     Cilk_mutex_signal(&USE_SHARED(dprintf_lock));
}

DECLARE_SHARED_VARIABLE(static Cilk_mutex, die_lock, );

void Cilk_die(const char *fmt,...)
{
     va_list l;

     Cilk_mutex_wait(&USE_SHARED(dprintf_lock));
     if (Self >= 0)
	  fprintf(stderr, "%d: fatal error: ", Self);
     else
	  fprintf(stderr, "global: fatal error: ");
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     va_end(l);
     Cilk_mutex_signal(&USE_SHARED(dprintf_lock));

     /* ensure only the first fatal error produces death of the program */
     if (Cilk_mutex_try(&USE_SHARED(die_lock)))
	  abort();		/* dump core */
}

/***********************************************************\
 * Mapping of variable segments
\***********************************************************/

#if !CILK_MANAGE_SHARED_SEGMENT 
/*
 * if linker segments are not used, a portable version
 * of these two functions is provided in cilk.c
 */

/* 
 * map the global variables segment as shared
 */
void Cilk_map_shared_segment(void)
{
     extern _end_Cilk_shared;	/* linker_defined symbols */
     extern _begin_Cilk_shared;
     char name[1024];
     int fd;
     void *startmem, *endmem;

     strcpy(name, "/tmp/_Cilk_gvars_XXXXXX");
     mktemp(name);

     startmem = &_begin_Cilk_shared;
     endmem = &_end_Cilk_shared;

     /* write all shared data to the file */
     fd = open(name, O_RDWR | O_CREAT, 0600);
     CILK_CHECK(fd >= 0, ("Open error\n"));
     unlink(name);
     CILK_CHECK(write(fd, startmem, endmem - startmem) == endmem - startmem,
		("error in write\n"));

     /* map shared data to be shared */
     if (startmem != mmap(startmem, endmem - startmem, PROT_READ | PROT_WRITE,
			  MAP_SHARED | MAP_FIXED, fd, 0))
	  Cilk_die("couldn't map shared part to file.\n");
     
     close(fd);
}
#endif

#if !CILK_MANAGE_PRIVATE_SEGMENT 
void Cilk_map_private_segment(void)
{
     /* variables are private by default */
}
#endif

/***********************************************************\
 * Malloc support
\***********************************************************/
/* Get a chunk of memory of size <len> that is mapped shared. */
void *Cilk_map_shared(size_t len)
{
     void *area;

     area = mmap(0, len, PROT_READ | PROT_WRITE, 
		 MAP_ANONYMOUS | MAP_SHARED, -1, 0);

     /* printf("MMap called %p len=%lx\n", area, len); */

     if (((long) area) == -1) {
	  /* 
	   * can't use Cilk_die, because the mutex in
	   * not initialized yet
	   */
	  fprintf(stderr, "mmap error in Cilk_map_shared -- "
		   "couldn't allocate %0f bytes\n", (double) len);
	  exit(1);
     }
     return area;
}

/***********************************************************\
 * Architecture-specific options
\***********************************************************/

/*
 * Parse any architecture-specific command-line options.  <argc>
 * and <argv> are the original arguments passed on the command line,
 * and <i> is the index of the first architecture-specific option (if any).
 * This function returns the index of the first user option, or <argc> if
 * there aren't any.
 */
int Cilk_arch_specific_options(int i, int argc, char *argv[])
{
  /* For now only support memory locks */
     USE_SHARED_RO(lock_type) = MEMORY_LOCKS;

     return i;
}

/***********************************************************\
 * Timer support
\***********************************************************/

DECLARE_PRIVATE_VARIABLE(static Cilk_time, Cilk_last_clock_low32, );
DECLARE_PRIVATE_VARIABLE(static Cilk_time, Cilk_last_clock_high32, );

/* Note: this returns a virtual time modulo 2^32.  Since we are
 * counting at a rate of 500MHz, the counter overflows in only 8
 * seconds!  This is not good enough for real life, but for now it is
 * passable.  Critical path measurements will work if all thread
 * lengths are small, because time is still accumulated in a 64-bit
 * value.
 */
Cilk_time Cilk_get_time(void)
{
  register Cilk_time clock_low32;
  register Cilk_time tmp;
  
  /* Read low 32 bits of clock.
   * From page 4-142 in the Alpha Architecture Handbook
   * http://ftp.digital.com/pub/Digital/info/semiconductor/literature/alphahb2.pdf
   */
  asm ("rpcc %0\n"
       "sll %0, 32, %1\n"
       "addq %0, %1, %0\n"
       "srl %0, 32, %0" : "=r" (clock_low32), "=r" (tmp) : );
  
  /* if we have wrapped around, add to high 32 bits of clock */
  if (clock_low32 < USE_PRIVATE(Cilk_last_clock_low32))
    {
      /* overflow occurred - fix up */
      USE_PRIVATE(Cilk_last_clock_high32) += (1LL<<32);
    }
  
  /* save most up-to-date low 32 bits of clock */
  USE_PRIVATE(Cilk_last_clock_low32) = clock_low32;
  
  /* return clock value */
  return USE_PRIVATE(Cilk_last_clock_high32) + clock_low32;
}

/***********************************************************\
 * Support for mutexen 
\***********************************************************/
/* #include <asm/system.h> */

/* initialize lock */
void Cilk_mutex_init_memory(Cilk_mutex *lock)
{
     lock->memory = 0;
     CILK_W_SYNC;
}

/*
 *  Obtain lock <lock>.  Waits until lock is available.
 */
void Cilk_mutex_wait_memory(Cilk_mutex *lock)
{
     int tmp;
     Cilk_enter_state(STATE_WAITING_FOR_LOCK);
     do {
          do {
               asm volatile ("ldl_l %0, (%1)":"=r" (tmp):"r"(&lock->memory));
          } while (tmp != 0);

          tmp = 1;
          asm volatile ("stl_c %0, (%1)":"=r" (tmp):"r"(&lock->memory), "0"(tmp));
     } while (tmp == 0);
     CILK_SYNC;
     Cilk_exit_state(STATE_WAITING_FOR_LOCK);
}

/*
 *  Try to obtain lock <lock>.  Returns 1 if it succeeds and 0 if it fails.
 *  Does not wait.
 */
int Cilk_mutex_try_memory(Cilk_mutex *lock)
{
     int tmp;

     asm volatile ("ldl_l %0, (%1)":"=r" (tmp):"r"(&lock->memory));
     if (tmp != 0)
          return 0;

     tmp = 1;
     asm volatile ("stl_c %0, (%1)":"=r" (tmp):"r"(&lock->memory), "0"(tmp));
     CILK_SYNC;
     return tmp;
}

/*
 *  Give up lock <lock>.
 */
void Cilk_mutex_signal_memory(Cilk_mutex *lock)
{
     CILK_SYNC;
     lock->memory = 0;
}



#if 0

/*******************************
 * Operating System locks
 *******************************/
/* initialize lock */
static inline void Cilk_mutex_init_os(Cilk_mutex *lock)
{
    pthread_mutex_init(&lock->os, NULL);
}

/*
 *  Obtain lock <lock>.  Waits until lock is available.
 */
static inline void Cilk_mutex_wait_os(Cilk_mutex *lock)
{
     Cilk_enter_state(STATE_WAITING_FOR_LOCK);

     pthread_mutex_lock(&lock->os);

     Cilk_exit_state(STATE_WAITING_FOR_LOCK);
}

/*
 *  Give up lock <lock>.
 */
static inline void Cilk_mutex_signal_os(Cilk_mutex *lock)
{
     pthread_mutex_unlock(&lock->os);
}

/*
 *  Try to obtain lock <lock>.  Returns 1 if it succeeds and 0 if it fails.
 *  Does not wait.
 */
static inline int Cilk_mutex_try_os(Cilk_mutex *lock)
{
     if (pthread_mutex_trylock(&lock->os) == 0) {
	  return 1;
     } else {
	  return 0;
     }
}
#endif



/*********************
 * Lock glue 
 *********************/

#if 0
#define MAKE_LOCKFUN(name) \
void name(Cilk_mutex *lock)                                   \
{                                                             \
     if (USE_SHARED_RO(lock_type) == MEMORY_LOCKS) {	      \
	  name##_memory(lock);				      \
     } else {						      \
	  name##_os(lock);				      \
     }							      \
}							      \

#else

#define MAKE_LOCKFUN(ret_type, name) \
ret_type name(Cilk_mutex *lock)                                   \
{                                                             \
	  name##_memory(lock);				      \
}

#endif

MAKE_LOCKFUN(void, Cilk_mutex_init)
MAKE_LOCKFUN(void, Cilk_mutex_wait)
MAKE_LOCKFUN(void, Cilk_mutex_signal)

int Cilk_mutex_try(Cilk_mutex *lock)
{
	  return Cilk_mutex_try_memory(lock);
}


							       
/***********************************************************\
 * Spawning worker processes
\***********************************************************/

#include <sys/wait.h>
#include <malloc.h>
#include <signal.h>
#include <sys/resource.h>

/*
 * Start forking processes: one process enters this function,
 * <Cilk_active_size> processes should exit this function immediately
 * and return 0.  Each returning process should set <Self> to a unique
 * number between 0 and <Cilk_active_size> - 1.  
 */
void Cilk_create_children(void (*child)(int))
{
     int i, n;
     pid_t pid;
     int stat;
     pid_t *children_pid;

     children_pid = (pid_t *) malloc(USE_SHARED_RO(active_size) 
				     * sizeof(pid_t));
     CILK_CHECK(children_pid,
		("couldn't malloc() children_pid\n"));

     for (i = 0; i < USE_SHARED_RO(active_size); i++) {
	  if ((pid = fork()) < 0) {
	       Cilk_die("can't fork");
	  } else if (pid == 0) {	/* child */
	       child(i);
	       exit(0);
	  } else {		/* parent */
	       children_pid[i] = pid;
	  }
     }

     /* parent now waits for all children to complete */
     for (n = 0; n < USE_SHARED_RO(active_size); n++) {
	  pid = waitpid((pid_t) - 1, &stat, 0);

	  /* find completed child's # */
	  for (i = 0; i < USE_SHARED_RO(active_size); i++)
	       if (pid == children_pid[i])
		    break;
	  CILK_ASSERT(i < USE_SHARED_RO(active_size));

	  if (WIFEXITED(stat)) {
	       if (WEXITSTATUS(stat) == 0) {
		    continue;
	       } else
		    Cilk_dprintf("%d: Child exited with non-zero value %d\n",
				 i, WEXITSTATUS(stat));
	  } else if (WIFSIGNALED(stat)) {
	       const char *core;
	       if (WCOREDUMP(stat)) {
		    core = " (core dumped)";
		    /* prevent others from overwriting core, if possible */
		    Cilk_mutex_try(&USE_SHARED(die_lock));
	       } else
		    core = "";

	       /* 
	        * Print out understandable messages for common signals,
	        * and the signal number for the uncommon ones. 
	        */
	       switch (WTERMSIG(stat)) {
		   case SIGILL:
			Cilk_dprintf("%d: Illegal instruction%s\n", i, core);
			break;
		   case SIGABRT:
			Cilk_dprintf("%d: Abort or Assertion failed%s\n", 
				     i, core);
			break;
		   case SIGFPE:
			Cilk_dprintf("%d: Floating exception%s\n", i, core);
			break;
		   case SIGKILL:
			Cilk_dprintf("%d: Terminated%s\n", i, core);
			break;
		   case SIGBUS:
			Cilk_dprintf("%d: Bus error%s\n", i, core);
			break;
		   case SIGSEGV:
			Cilk_dprintf("%d: Segmentation fault%s\n", i, core);
			break;
		   default:
			Cilk_dprintf(
			  "%d: signal %d (see /usr/include/sys/signal.h)"
					 "%s\n", i, WTERMSIG(stat), core);
			break;
	       }
	  }
	  /* 
	   * something went wrong with child i: abort the rest of the 
	   * children.  Sending a SIGKILL keeps them from dumping core.
	   */
	  for (i = 0; i < USE_SHARED_RO(active_size); i++)
	       kill(children_pid[i], SIGKILL);
     }
}


/**********************************************************
 * Priority control
 **********************************************************/

void Cilk_lower_priority(void)
{
     usleep(2000);
}

void Cilk_raise_priority(void)
{
}

void Cilk_raise_all_priority(void)
{
}

/**********************************************************
 * Initialization of the module
 **********************************************************/

void Cilk_arch_specific_init(void)
{
     Cilk_mutex_init(&USE_SHARED(dprintf_lock));
     Cilk_mutex_init(&USE_SHARED(die_lock));
}

void Cilk_arch_specific_per_process_init(void)
{
     USE_PRIVATE(Cilk_last_clock_low32) = 0;
     USE_PRIVATE(Cilk_last_clock_high32) = 0;
}

