#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>
#include <cilk-cilk2c.h>

/* Check the defines */
#ifndef CILK_LIBRARY_VERSION
#error The library version of Cilk must be compiled with CILK_LIBRARY_VERSION defined
#endif
#ifndef CILK_LIBRARY_VERSION_SUPPORTED
#error This architecture file does not support the library version of Cilk
#endif

static Closure *create_initial_foo_thread(int (*foo)(void*), void *arg);

void Cilk_init(Cilk_options *options)
{
     /* initialize global part of system */
     Cilk_global_init(options);

     /* initialize the scheduler */
     Cilk_scheduler_init();

     Cilk_library_init();

     /* create children */
     Cilk_create_children(Cilk_child_main);
}


void Cilk_finish(void)
{
  Cilk_library_exit();
}

/* save top-level closure for use by Cilk_exit() */
DECLARE_SHARED_VARIABLE(static struct Closure *, top_level, );

int Cilk_run(int (*foo)(void*), void *arg)
{
  Closure *t;
  
  Cilk_library_ready_to_run();

  /*
   * make initial closure
   */
  t = create_initial_foo_thread(foo, arg);
  USE_SHARED(top_level) = t;
  
  /*
   * install initial closure in first worker's queue so it will get executed
   */
  Cilk_install_closure(t, 0);

  Cilk_fence();

  /* wake up the library and run the computation */
  Cilk_library_run();

  /* wait until the computation terminates */
  Cilk_library_main_sleep();

  /*
   * Make sure all closures are done
   */
  Cilk_assert_empty_deques();

  /*
   * Return foo's result
   */
  return USE_SHARED(retval);
}


/************************************************************
 *  Initial thread
 ************************************************************/

/* This is a hand-compiled procedure that calls foo procedure */

struct invoke_foo_catch_inlet_args {int res;};

struct invoke_foo_frame {
     StackFrame header;
     int (*foo)(void*);
     void *arg;
     struct invoke_foo_catch_inlet_args inlet_args;
};

/* you are not supposed to understand this */


/* 
 * This inlet is created for the purpose of Cilk_exit(). 
 * It sets retval to the exit value, and calls abort.
 * So, when this inlet is excuted, the user's Cilk_run
 * should return the argument of Cilk_exit().
 */
static void invoke_foo_catch_inlet(struct invoke_foo_frame *_frame,
				    struct invoke_foo_catch_inlet_args *inletargs,
				    void* inletresult)
{
     USE_SHARED(retval) = inletargs->res;
     _ABORT_STANDALONE();
}

static void invoke_foo_slow(struct invoke_foo_frame *_frame)
{
     int (*foo)(void*);
     void *arg;
     int tmp;

     _START_THREAD_SLOW();
     switch (_frame->header.entry) {
     case 1: 
	  goto _sync1;
     case 2: 
	  goto _sync2;
     }
     
     
     Cilk_user_work = 0;
     Cilk_user_critical_path = 0;
     WHEN_CILK_TIMING(_frame->header.cp = (Cilk_time) 0);
     WHEN_CILK_TIMING(_frame->header.work = (Cilk_time) 0);
     WHEN_CILK_TIMING(_frame->header.mycp = (Cilk_time) 0);
     WHEN_CILK_TIMING(USE_PRIVATE(last_cp_time) = Cilk_get_time());

     foo = _frame->foo;
     arg = _frame->arg;

     _frame->header.index = (char *) &_frame->inlet_args -
	  (char *) _frame;
     _frame->header.entry = 1;
     _BEFORE_SPAWN_SLOW();
     _PUSH_FRAME(_frame);
     tmp = foo(arg);
     _XPOP_FRAME_RESULT(_frame,/* return nothing */,tmp);
     USE_SHARED(retval) = tmp;
     _AFTER_SPAWN_SLOW();

     if (0) {
     _sync1:
	  ;
     }
     _AT_THREAD_BOUNDARY_SLOW();
     
     _BEFORE_SYNC_SLOW();
     if (_SYNC) {
          _frame->header.entry=2;
	  return;
     _sync2:
	  ;
     }
     _AFTER_SYNC_SLOW();
     _AT_THREAD_BOUNDARY_SLOW();

     WHEN_CILK_TIMING(USE_SHARED(critical_path) = _frame->header.mycp);
     WHEN_CILK_TIMING(USE_SHARED(total_work) = _frame->header.work);
     
     Cilk_remove_and_free_closure_and_frame(&_frame->header, Self);
     
     Cilk_library_main_wakeup();
}


static ProcInfo invoke_foo_sig[]={
     {sizeof(int), sizeof(struct invoke_foo_frame ), invoke_foo_slow, 0, 0},
     {sizeof(int), 0, invoke_foo_catch_inlet,
      sizeof(struct invoke_foo_catch_inlet_args),
      offsetof(struct invoke_foo_frame, inlet_args)},
     {0, 0, 0, 0, 0}
};


static Closure *create_initial_foo_thread(int (*foo)(void*), void *arg)
{
     Closure *t;
     struct invoke_foo_frame *f;

     /* create a frame for invoke_foo */
     t = Cilk_Closure_create_malloc();
     t->parent = (Closure *) NULL;
     t->status = CLOSURE_READY;

     f = Cilk_malloc(sizeof(struct invoke_foo_frame));
     f->header.entry = 0;
     f->header.sig = invoke_foo_sig;
     WHEN_CILK_DEBUG(f->header.magic = CILK_STACKFRAME_MAGIC);

     f->foo = foo;
     f->arg = arg;

     t->frame = &f->header;
     return t;
}

void Cilk_exit_from_user_main(Closure *cl, int res);
void Cilk_exit(int res)
{
  Cilk_exit_from_user_main(USE_SHARED(top_level), res);
}

