/*
 * $Id: sched.c,v 1.9 1998/05/21 21:40:41 lisiecki Exp $
 */

/*
 * $Log: sched.c,v $
 * Revision 1.9  1998/05/21 21:40:41  lisiecki
 * Library version support
 *
 * Revision 1.8  1997/11/18 19:12:19  randall
 * Fix to reduce memory bandwidth consumed by stealers.
 *
 * Revision 1.7  1997/10/01 18:50:46  randall
 * Final fix.
 *
 * Revision 1.6  1997/10/01 18:48:33  randall
 * Fixed 64-bit random number generator bug.
 *
 * Revision 1.5  1997/09/17 19:07:34  randall
 * Added digital unix thread version.
 *
 * Revision 1.4  1997/09/04 19:12:35  athena
 * Fixed error messages to be more personal and polite to the user.
 *
 * Revision 1.3  1997/09/03 23:00:57  bsong
 * Added some helpful error message for the user by using CILK_COMPLAIN.
 *
 * Revision 1.2  1997/03/01 17:42:56  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/*
 * This file contains the scheduler and associated functions
 */

#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>
#include <string.h>

DECLARE_SHARED_VARIABLE(int, retval,);	/* return value from main */
DECLARE_SHARED_VARIABLE(int, done,);
#ifdef CILK_LIBRARY_VERSION
DECLARE_SHARED_VARIABLE(int, exit,);   /* set by Cilk_exit */
#endif

/* 
 * The ready deque.  Each ready queue contains a list of ready closures;
 * the last element of the deque may also be RUNNING or RETURNING
 */
typedef struct ReadyDeque {
     Cilk_mutex mutex;
     WHEN_CILK_DEBUG(int mutex_owner;)
     Closure *top, *bottom;
     CILK_CACHE_LINE_PAD;
} ReadyDeque;


DECLARE_PRIVATE_VARIABLE(ClosureCache *, cache ,);
DECLARE_PRIVATE_VARIABLE(int, _Self, );

WHEN_CILK_TIMING(DECLARE_PRIVATE_VARIABLE(Cilk_time, last_cp_time,);)
WHEN_CILK_TIMING(DECLARE_PRIVATE_VARIABLE(Cilk_time, cp_hack,);)
WHEN_CILK_TIMING(DECLARE_PRIVATE_VARIABLE(Cilk_time, work_hack,);)

DECLARE_PRIVATE_VARIABLE(Cilk_time, user_work,);
DECLARE_PRIVATE_VARIABLE(Cilk_time, user_critical_path,);

#define CLOSURE_HEAD(cl)  (cl->cache->head)
#define CLOSURE_EXCEPTION(cl)  (cl->cache->exception)
#define CLOSURE_TAIL(cl)  (cl->cache->tail)
#define CLOSURE_STACK(cl) (cl->cache->stack)
#define CLOSURE_ABORT(cl) (cl->abort_status)

#define NOBODY (-1)   /* invalid processor number */

/* 
 * SCHEDULER LOCK DIAGRAM
 * 
 * locks must be grabbed in the following order:
 *
 * ReadyDeque lock
 * Closure lock
 *
 * Closure locks must be acquired downwards
 * (parent first, then children)
 *
 * See also PROTOCOLS for a more verbose description.
 */

/*************************************************************
 * Basic operations on closures
 *************************************************************/
static inline void Closure_checkmagic(Closure *t)
{
     CILK_ASSERT(t->magic == CILK_CLOSURE_MAGIC);
}

/* assert that t be locked by ourselves */
static inline void Closure_assert_ownership(Closure *t)
{
     CILK_ASSERT(t->mutex_owner == Self);
}

/* assert that t be not locked */
static inline void Closure_assert_alienation(Closure *t)
{
     CILK_ASSERT(t->mutex_owner != Self);
}

static inline void Closure_destroy(Closure *t)
{
     Closure_checkmagic(t);

     WHEN_CILK_DEBUG(t->magic = ~CILK_CLOSURE_MAGIC);
     CILK_ASSERT(! (t->malloced));
     Cilk_internal_free(t, sizeof(Closure));
}

static inline void Closure_destroy_malloc(Closure *t)
{
     Closure_checkmagic(t);

     WHEN_CILK_DEBUG(t->magic = ~CILK_CLOSURE_MAGIC);
     CILK_ASSERT((t->malloced));
     Cilk_free(t);
}

static inline void Closure_lock(Closure *t)
{
     Closure_checkmagic(t);
     Cilk_mutex_wait(&(t->mutex));
     WHEN_CILK_DEBUG(t->mutex_owner = Self);
}

static inline void Closure_unlock(Closure *t)
{
     Closure_checkmagic(t);
     Closure_assert_ownership(t);
     WHEN_CILK_DEBUG(t->mutex_owner = -1);
     Cilk_mutex_signal(&(t->mutex));
}

/***********************************************************
 *  Internal random number generator, since rand()/random() is not
 *  MT-Safe. This RNG is *distinct* from the user RNG, just 
 *  for safety.
 ***********************************************************/
/* code stolen from NetBSD */

/*
 * Cilk_rand_next is stored in the per-processor global section to
 * make this code MT-Safe.
 */
DECLARE_PRIVATE_VARIABLE(static unsigned int, rand_next, );

static unsigned int rts_rand(void)
{
     USE_PRIVATE(rand_next) = USE_PRIVATE(rand_next) * 1103515245 + 12345;
     return (USE_PRIVATE(rand_next) >> 16);
}

static void rts_srand(unsigned int seed)
{
     USE_PRIVATE(rand_next) = seed;
}

/*********************************************************
 * Management of ReadDeques
 *********************************************************/

/* dynamically-allocated array of deques, one per processor */
DECLARE_SHARED_VARIABLE(static struct ReadyDeque *, deques, );

/* initialize the deques */
static void init_deques(void)
{
     int i;

     CILK_ASSERT(Self == NOBODY);
     USE_SHARED(deques)= Cilk_malloc_fixed(USE_SHARED_RO(active_size) *
					   sizeof(ReadyDeque));
     CILK_CHECK(USE_SHARED(deques), ("failed to allocate deques\n"));

     for (i = 0; i < USE_SHARED_RO(active_size); ++i) {
	  USE_SHARED(deques)[i].top = (Closure *) NULL;
	  USE_SHARED(deques)[i].bottom = (Closure *) NULL;
	  Cilk_mutex_init(&USE_SHARED(deques)[i].mutex);
	  WHEN_CILK_DEBUG(USE_SHARED(deques)[i].mutex_owner = NOBODY);
     }
}

/* assert that pn's deque be locked by ourselves */
static inline void deque_assert_ownership(int pn)
{
     CILK_ASSERT(USE_SHARED(deques)[pn].mutex_owner == Self);
}

static inline void deque_assert_alienation(int pn)
{
     CILK_ASSERT(USE_SHARED(deques)[pn].mutex_owner != Self);
}

static inline void deque_lock(int pn)
{
     Cilk_mutex_wait(&USE_SHARED(deques)[pn].mutex);
     WHEN_CILK_DEBUG(USE_SHARED(deques)[pn].mutex_owner = Self);
}

static inline void deque_unlock(int pn)
{
     WHEN_CILK_DEBUG(USE_SHARED(deques)[pn].mutex_owner = NOBODY);
     Cilk_mutex_signal(&USE_SHARED(deques)[pn].mutex);
}

/* 
 * functions that add/remove elements from the top/bottom
 * of deques
 */
static Closure *deque_xtract_top(int pn)
{
     Closure *cl;

     deque_assert_ownership(pn);

     cl = USE_SHARED(deques)[pn].top;
     if (cl) {
	  CILK_ASSERT(cl->owner_ready_deque == pn);
	  USE_SHARED(deques)[pn].top = cl->next_ready;
	  if (cl == USE_SHARED(deques)[pn].bottom) {
	       CILK_ASSERT(cl->next_ready == (Closure *) NULL);
	       USE_SHARED(deques)[pn].bottom = (Closure *) NULL;
	  } else {
	       CILK_ASSERT(cl->next_ready);
	       (cl->next_ready)->prev_ready = (Closure *) NULL;
	  }
	  WHEN_CILK_DEBUG(cl->owner_ready_deque = NOBODY);
     } else {
	  CILK_ASSERT(USE_SHARED(deques)[pn].bottom == (Closure *) NULL);
     }
     
     return cl;
}

static Closure *deque_peek_top(int pn)
{
     Closure *cl;

     deque_assert_ownership(pn);

     cl = USE_SHARED(deques)[pn].top;
     if (cl) {
	  CILK_ASSERT(cl->owner_ready_deque == pn);
     } else {
	  CILK_ASSERT(USE_SHARED(deques)[pn].bottom == (Closure *) NULL);
     }

     return cl;
}

static Closure *deque_xtract_bottom(int pn)
{
     Closure *cl;

     deque_assert_ownership(pn);

     cl = USE_SHARED(deques)[pn].bottom;
     if (cl) {
	  CILK_ASSERT(cl->owner_ready_deque == pn);
	  USE_SHARED(deques)[pn].bottom = cl->prev_ready;
	  if (cl == USE_SHARED(deques)[pn].top) {
	       CILK_ASSERT(cl->prev_ready == (Closure *) NULL);
	       USE_SHARED(deques)[pn].top = (Closure *) NULL;
	  } else {
	       CILK_ASSERT(cl->prev_ready);
	       (cl->prev_ready)->next_ready = (Closure *) NULL;
	  }

	  WHEN_CILK_DEBUG(cl->owner_ready_deque = NOBODY);
     } else {
	  CILK_ASSERT(USE_SHARED(deques)[pn].top == (Closure *) NULL);
     }
     
     return cl;
}

static Closure *deque_peek_bottom(int pn)
{
     Closure *cl;

     deque_assert_ownership(pn);

     cl = USE_SHARED(deques)[pn].bottom;
     if (cl) {
	  CILK_ASSERT(cl->owner_ready_deque == pn);
     } else {
	  CILK_ASSERT(USE_SHARED(deques)[pn].top == (Closure *) NULL);
     }
     
     return cl;
}

#if 0   
/*
 * Unused code, that may be useful if you hack the scheduler.
 */
static void deque_add_top(Closure *cl, int pn)
{
     deque_assert_ownership(pn);
     CILK_ASSERT(cl->owner_ready_deque == NOBODY);

     cl->next_ready = USE_SHARED(deques)[pn].top;
     cl->prev_ready = (Closure *)NULL;
     USE_SHARED(deques)[pn].top = cl;
     WHEN_CILK_DEBUG(cl->owner_ready_deque = pn);
     
     if (USE_SHARED(deques)[pn].bottom) {
	  CILK_ASSERT(cl->next_ready);
	  (cl->next_ready)->prev_ready = cl;
     } else {
	  USE_SHARED(deques)[pn].bottom = cl;
     }
}
#endif

static void deque_add_bottom(Closure *cl, int pn)
{
     deque_assert_ownership(pn);
     CILK_ASSERT(cl->owner_ready_deque == NOBODY);

     cl->prev_ready = USE_SHARED(deques)[pn].bottom;
     cl->next_ready = (Closure *)NULL;
     USE_SHARED(deques)[pn].bottom = cl;
     WHEN_CILK_DEBUG(cl->owner_ready_deque = pn);
     
     if (USE_SHARED(deques)[pn].top) {
	  CILK_ASSERT(cl->prev_ready);
	  (cl->prev_ready)->next_ready = cl;
     } else {
	  USE_SHARED(deques)[pn].top = cl;
     }
}

static inline void deque_assert_is_bottom(Closure *t)
{
     deque_assert_ownership(Self);
     CILK_ASSERT(t == deque_peek_bottom(Self));
}

#ifdef CILK_LIBRARY_VERSION
/* Install closure t in pn's deque */
void Cilk_install_closure(Closure *t, int pn) {
  CILK_ASSERT(pn >= 0 && pn < USE_SHARED_RO(active_size));

  deque_lock(pn);
  deque_add_bottom(t, pn);
  deque_unlock(pn);
}

/* Assert that all deques are empty */
void Cilk_assert_empty_deques() {
  int i;
  for(i=0; i<USE_SHARED_RO(active_size); i++)
    {
      deque_lock(i);
      CILK_ASSERT(deque_peek_top(i) == NULL);
      deque_unlock(i);
    }
}
#endif /* ifdef CILK_LIBRARY_VERSION */

/* Remove closure for frame f from bottom of pn's deque and free them */
void Cilk_remove_and_free_closure_and_frame(StackFrame *f, int pn) {
  Closure *t;

  deque_lock(pn);
  t = deque_xtract_bottom(pn);
  CILK_ASSERT(t->frame == f);
  deque_unlock(pn);
  Cilk_free(f);
  Closure_destroy_malloc(t);
}

/*********************************************************
 * Management of ClosureCache's
 *********************************************************/

/* dynamically-allocated array of ClosureCache's, one per processor */
DECLARE_SHARED_RO_VARIABLE(static ClosureCache *, caches, );

/* initialize the caches */
static void init_caches(void)
{
     int i;

     CILK_ASSERT(Self == NOBODY);
     INIT_SHARED_RO(caches,
	  Cilk_malloc_fixed(USE_SHARED_RO(active_size) * 
			    sizeof(ClosureCache)));
     CILK_CHECK(USE_SHARED_RO(caches), ("failed to allocate caches\n"));

     for (i = 0; i < USE_SHARED_RO(active_size); ++i) {
	  USE_SHARED_RO(caches)[i].stack =
	       Cilk_malloc_fixed(USE_SHARED_RO(stackdepth) *
				 sizeof(StackFrame *));
	  CILK_CHECK(USE_SHARED_RO(caches)[i].stack,
		     ("failed to allocate stack\n"));
     }
}

static inline ClosureCache *cache_get_addr(int pn)
{
     return USE_SHARED_RO(caches) + pn;
}

/**************************************************
 * Management of ProcInfo's
 **************************************************/
/*
 * Returns a pointer to the ProcInfoT structure describing entry <entry>
 * of procedure whose signature is p.
 */
static inline ProcInfo *get_ProcInfo(ProcInfo * p, int entry)
{
     CILK_ASSERT(entry > 0);
     return p + entry;
}

/*
 * Returns a pointer to the slow version for a procedure
 * whose signature is p.
 */
static void (*get_proc_slow(ProcInfo *p)) () {
     return p[0].inlet;
}

/********************************************
 * Closure management
 ********************************************/
static inline int Closure_at_top_of_stack(Closure *cl)
{
     return (CLOSURE_HEAD(cl) + 1 == CLOSURE_TAIL(cl));
}

static inline int Closure_has_children(Closure *cl)
{
     return (cl->join_counter != 0);
}

static inline void Closure_init(Closure *new)
{
     Cilk_mutex_init(&new->mutex);

     new->cp = (Cilk_time) 0;
     new->work = (Cilk_time) 0;

     new->complete_inlets = (struct InletClosure *) NULL;
     new->incomplete_inlets = (struct InletClosure *) NULL;
     new->next_ready = (struct Closure *) NULL;
     new->prev_ready = (struct Closure *) NULL;
     CLOSURE_ABORT(new) = NO_ABORT;
     new->return_size = 0;
     new->return_value = (void *)0;

     WHEN_CILK_DEBUG({
	  new->magic = CILK_CLOSURE_MAGIC;
	  new->owner_ready_deque = NOBODY;
	  new->mutex_owner = NOBODY;
     });

     Cilk_event(EVENT_CLOSURE_CREATE);
}

static Closure *Closure_create(void)
{
     Closure *new = Cilk_internal_malloc(sizeof(Closure));

     CILK_CHECK(new, ("can't allocate Closure\n"));
     WHEN_CILK_DEBUG(new->malloced = 0;)

     Closure_init(new);

     return new;
}

Closure *Cilk_Closure_create_malloc(void)
{
     Closure *new = Cilk_malloc(sizeof(Closure));

     CILK_CHECK(new, ("can't allocate Closure\n"));
     WHEN_CILK_DEBUG(new->malloced = 1;)

     Closure_init(new);

     return new;
}

/* 
 * if (entry >= 0)
 *      get the inlet according to the entry
 * else
 *      get the inlet from the frame
 */
static struct InletClosure *make_incomplete_inlet_closure(Closure *parent, 
							  int entry)
{
     struct InletClosure *p;
     StackFrame *f;
     ProcInfo *info;

     p = (struct InletClosure *)
	 Cilk_internal_malloc(sizeof(struct InletClosure));

     f = parent->frame;

     if (entry >= 0) 
	  /* get the special exit closure */
	  info = get_ProcInfo(f->sig, entry);
     else
	  info = get_ProcInfo(f->sig, f->entry);
     
     /* if no index is statically specified, get it from the frame */
     if (info->index == -1)
	  p->index = f->index;
     else
          p->index = info->index;

     /* take care of inlets */
     if (info->inlet != NULL) {
	  /* get arguments of inlet out of frame and into closure */
          p->inlet = info->inlet;
          p->argsize = info->argsize;
	  p->inlet_args = Cilk_internal_malloc(info->argsize);

	  memcpy((void *) p->inlet_args,
		 (char *) f + info->argindex, info->argsize);
     } else {
	  /*
	   * in the non-inlet case, we use the inlet_args field
	   * to hold the return value
	   */
	  p->inlet = NULL;
	  p->argsize = info->size;
	  p->inlet_args = Cilk_internal_malloc(info->size);
     }
     
     return p;
}

/* 
 * only the scheduler is allowed to alter the closure
 * tree.  Consequently, these operations are private
 */
static void Closure_add_child(Closure *parent,
			      Closure *child)
{
     struct InletClosure *p;

     Closure_checkmagic(parent);
     Closure_assert_ownership(parent);
     Closure_checkmagic(child);
     Closure_assert_alienation(child);

     p = make_incomplete_inlet_closure(parent, -1);

     /*
      * semantic overloading: the inlet closure also contains
      * the list of children.
      */
     p->this = child;

     /* splice into list of unreturned children in parent */
     p->next = parent->incomplete_inlets;
     parent->incomplete_inlets = p;
}

static struct InletClosure *Closure_remove_child(Closure *parent,
						 Closure *child)
{
     struct InletClosure **q;
     struct InletClosure *p;

     Closure_checkmagic(parent);
     Closure_checkmagic(child);
     Closure_assert_ownership(parent);

     for (q = &parent->incomplete_inlets; *q; q = &((*q)->next)) {
	  p = *q;
	  if (p->this == child) {
	       *q = p->next;
	       goto found;
	  }
     }

     Cilk_die("BUG in Closure_remove_child\n");
     p = NULL;

   found:
     return p;
}

/*******************************************
 * Exceptions
 *******************************************/
/*
 * Cilk supports a clever way to send an interruption/exception
 * to a running worker, without interrupting its normal way
 * of working. The idea is as follows.
 *
 * The local worker operates at the bottom of the frame
 * stack, using a stack pointer T (tail).  The exception
 * mechanism maintains another pointer E (exception).
 * The worker polls the value of E when decrementing T.
 * Whenever T becomes less or equal than E, the worker
 * calls the exception handler of the runtime system,
 * that does all the necessary bookkeeping.
 *
 * Exceptions can be sent only to running closures.
 * The exception pointer is always greater or equal than
 * the head pointer (the index of the first non-stolen
 * frame on the stack)
 */

/* value of E bigger than any possible value of T */
#define EXCEPTION_INFINITY (volatile StackFrame**)0xFFFFFFFF

static inline void increment_exception_pointer(Closure *cl)
{
     Closure_assert_ownership(cl);

     CILK_ASSERT(cl->status == CLOSURE_RUNNING);

     if (CLOSURE_EXCEPTION(cl) != EXCEPTION_INFINITY) {
	  ++CLOSURE_EXCEPTION(cl);
	  
	  /* make sure the exception is visible, before we continue */
	  Cilk_fence();
     }
}

static inline void decrement_exception_pointer(Closure *cl)
{
     Closure_assert_ownership(cl);

     CILK_ASSERT(cl->status == CLOSURE_RUNNING);

     if (CLOSURE_EXCEPTION(cl) != EXCEPTION_INFINITY)
	  --CLOSURE_EXCEPTION(cl);
}

static inline void reset_exception_pointer(Closure *cl)
{
     Closure_assert_ownership(cl);

     CLOSURE_EXCEPTION(cl) = CLOSURE_HEAD(cl);
}

static inline void signal_immediate_exception(Closure *cl)
{
     Closure_assert_ownership(cl);

     CILK_ASSERT(cl->status == CLOSURE_RUNNING);

     CLOSURE_EXCEPTION(cl) = EXCEPTION_INFINITY;

     /* make sure the exception is visible, before we continue */
     Cilk_fence();
}

/********************************************************
 * Abort-related stuff 
 ********************************************************/

/* 
 * Abort is implemented by adding an abort_status in the closure
 * structure.  
 *    NO_ABORT:         the closure is not aborted;
 *    ALMOST_NO_ABORT:  the closure is being aborted as parent;
 *    ABORT_ALL:        the closure is being aborted as child;
 *
 * When abort happens, we want to make sure that already spawned
 * but not returned subcomputations stop working, and "return" as 
 * quickly as possible.  This is handled in the exception handler.
 * Also,  the inlets of a closure that is being aborted, either as
 * parent or child, should not be executed.  We handle this in
 * function apply_inlets.
 *
 * Abort is signaled either in running slow code, or in apply_inlets.
 * To signal an abort, the process that discovers the abort grabs
 * the parent lock and recursively goes down the tree to signal 
 * it's children by setting their abort_status to ABORT_ALL, and 
 * signal immediate exception to all running children.  The parent
 * is set to ALMOST_NO_ABORT, so that the complete inlets would not
 * be executed.  
 *
 * Before we signal an abort from the parent, we check to see that, if
 * the parent is running and the worker is not on the top of the stack,
 * then, we promote the frame second to the top to be a complete closure,
 * put the parent back to the ready deque, and signal abort from the parent.
 * In this way, the parent will start execution at the right entry point.
 * 
 * The stealing protocol does not steal a closure if it is being 
 * aborted.  Inlets are also not applied to closures that are aborting.
 *  
 * Once the abort is signaled, we let the normal protocol to handle
 * the abort:
 *
 *    We let the exception handler do the work to abort running closure.
 *    (see comments in exception_handler).
 *
 *    For a ready closure, we check whether it's being aborted before 
 *    set up to execute it.  If it is being aborted, we free the last 
 *    frame and let it return.
 *
 *    For a returning closure, if it is being aborted, we do not append 
 *    the corresponding inlet to the complete inlet list, but discard
 *    the return value.
 *
 *    For a suspended closure, we do the same provably_good_steal.  
 *    If it is being aborted, we will discover it we it's state is ready.  
 *    
 *
 */

DECLARE_PRIVATE_VARIABLE(int, abort_flag, );

static inline int get_frame_size(StackFrame *f) {
     return (f->sig)[0].index; /* compiler put the frame size here. */
}

static inline int Closure_being_aborted(Closure *cl)
{
     Closure_assert_ownership(cl);
     return (CLOSURE_ABORT(cl) != NO_ABORT);
}

static inline void signal_abort(Closure *cl, enum AbortStatus s)
{
     if (s < CLOSURE_ABORT(cl))
	  CLOSURE_ABORT(cl) = s;
}

static inline void reset_abort(Closure *cl)
{
     CLOSURE_ABORT(cl) = NO_ABORT;
}

static inline void maybe_reset_abort(Closure *cl)
{
     if (CLOSURE_ABORT(cl) != ABORT_ALL)
	  reset_abort(cl);
}


static void recursively_signal_abort(Closure *cl, enum AbortStatus s)
{
     struct InletClosure *p;

     Closure_assert_ownership(cl);
     signal_abort(cl, s);
     if (cl->status == CLOSURE_RUNNING)
	  signal_immediate_exception(cl);

     /* walk the list of children */
     for (p = cl->incomplete_inlets; p; p = p->next) {
	  Closure_lock(p->this);
	  recursively_signal_abort(p->this, ABORT_ALL);
	  Closure_unlock(p->this);
     }
}


static Closure *abort_ready_closure(Closure *cl)
{
     /*
      * abort a ready closure. There are two cases: if it has
      * children, suspend it.  Else free the frame associated
      * with the closure and make the closure return
      */
     CILK_ASSERT(CLOSURE_ABORT(cl) == ABORT_ALL);

     if (Closure_has_children(cl)) {
	  cl->status = CLOSURE_SUSPENDED;
	  Cilk_event(EVENT_ABORT_READY_SUSPEND);
	  return (Closure *)0;
     } else {
	  Cilk_destroy_frame(cl->frame, get_frame_size(cl->frame));
	  Cilk_event(EVENT_ABORT_READY_RETURN);
	  cl->status = CLOSURE_RETURNING;
	  return cl;
     }
}

/***********************************************************
 *         Work-stealing and related functions
 ***********************************************************/
/* 
 * For stealing we use a Dekker-like protocol, that achieves
 * mutual exclusion using shared memory.
 *
 * Thief:
 *   lock
 *   E++          ; signal intention to steal, sending an exception
 *   (implicit MEMBAR)
 *   If (H >= T)  ; if we can steal
 *     E--        ; retract the exception
 *     give up
 *   steal (that is, H++, etc)
 *   unlock
 *
 * The victim does the normal exception-handling mechanism:
 *
 *   T--
 *   --MEMBAR-- ensure T write occurs before E is read.
 *   if (E>=T)
 *     
 *
 */

/*
 * Do the thief part of Dekker's protocol.  Return 1 upon success,
 * 0 otherwise.  The protocol fails when the victim already popped
 * T so that E=T.
 */
static int do_dekker_on(Closure *cl)
{
     Closure_assert_ownership(cl);

     increment_exception_pointer(cl);
     Cilk_membar_StoreLoad(); 

     if ((CLOSURE_HEAD(cl) + 1) >= CLOSURE_TAIL(cl)) {
	  decrement_exception_pointer(cl);
	  return 0;
     }

     return 1;
}

static void Closure_make_ready(Closure *cl)
{
     cl->cache = (ClosureCache *)NULL;
     cl->status = CLOSURE_READY;
}

/*
 * promote the child frame of parent to a full closure.
 * Detach the parent and return it.
 *
 * Assumptions: the parent is running on victim, and we own
 * the locks of both parent and deque[victim].
 * The child keeps running on the same cache of the parent.
 * The parent's join counter is incremented.
 *
 * In order to promote a child frame to a closure,
 * the parent's frame must be the last in its ready queue.
 *
 * Returns the child.
 */
static Closure *promote_child(Closure *parent, int victim)
{
     Closure *child = Closure_create();

     Closure_assert_ownership(parent);
     deque_assert_ownership(victim);
     CILK_ASSERT(parent->status == CLOSURE_RUNNING);
     CILK_ASSERT(parent->owner_ready_deque == victim);
     CILK_ASSERT(parent->next_ready == (Closure *) NULL);

     /* can't promote until we are sure nobody works on the frame */
     CILK_ASSERT(CLOSURE_HEAD(parent) <= CLOSURE_EXCEPTION(parent));

     /* update the various fields */
     child->parent = parent;
     child->join_counter = 0;
     child->cache = parent->cache;
     child->status = CLOSURE_RUNNING;

     /* setup the frame of the child closure */
     CLOSURE_HEAD(child)++;
     child->frame = (StackFrame*)*CLOSURE_HEAD(child);

     /* insert the closure on the victim processor's deque */
     deque_add_bottom(child, victim);

     /* at this point the child can be freely executed */
     return child;
}

/*
 * Finishes the promotion process.  The child is already fully promoted
 * and requires no more work (we only use the given pointer to identify
 * the child).  This function does some more work on the parent to make
 * the promotion complete.
 */
void finish_promote(Closure *parent, Closure *child)
{
     Closure_assert_ownership(parent);
     Closure_assert_alienation(child);
     
     /* the parent is still locked; we still need to update it */
     /* join counter update */
     parent->join_counter++;

     /* register this child */
     Closure_add_child(parent, child);

     /* Make the parent ready (also invalidates the cache) */
     Closure_make_ready(parent);

     return;
}

/*
 * stealing protocol.  Tries to steal from the victim; returns a
 * stolen closure, or NULL if none.
 */
static Closure *Closure_steal(int victim)
{
     Closure *res = (Closure *) NULL;
     Closure *cl, *child;

     /* first attempt to see if any steal is possible.  If there
      * are one or zero closures in the deque, steal is not possible.
      */
     {
       ClosureCache *cache = cache_get_addr(victim);
       if (cache->head >= cache->tail - 1)
	 return res;
     }

     Cilk_event(EVENT_STEAL_ATTEMPT);

     deque_lock(victim);

     cl = deque_peek_top(victim);

     if (cl) {
	  Closure_lock(cl);
	  
	  /* do not steal aborting closures */
	  if (Closure_being_aborted(cl)) {
	       Cilk_event(EVENT_STEAL_ABORT);
	       goto steal_give_up;
	  }

	  switch (cl->status) {
	      case CLOSURE_READY:
		   Cilk_event(EVENT_STEAL);
		   res = deque_xtract_top(victim);
		   CILK_ASSERT(cl == res);

		   /* it is ok to swap these two */
		   Closure_unlock(cl);
		   deque_unlock(victim);
		   break;

	      case CLOSURE_RUNNING:
		   /* send the exception to the worker */
		   if (do_dekker_on(cl)) {
			/* 
			 * if we could send the exception, promote
			 * the child to a full closure, and steal
			 * the parent
			 */
		        child = promote_child(cl, victim);

			/* detach the parent */
			res = deque_xtract_top(victim);
			CILK_ASSERT(cl == res);
			deque_unlock(victim);

			/* 
			 * at this point, more steals can happen
			 * from the victim.
			 */
			finish_promote(cl, child);
		        Closure_unlock(cl);
			Cilk_event(EVENT_STEAL);
		   } else {
			/* it is ok to swap these two */
			Cilk_event(EVENT_STEAL_NO_DEKKER);
			goto steal_give_up;
		   }
		   break;
			
	      case CLOSURE_SUSPENDED:
		   Cilk_die("Bug: suspended closure in ready deque\n");
		   break;

	      case CLOSURE_RETURNING:
		   /* ok, let it leave alone */
		   Cilk_event(EVENT_STEAL_RETURNING);

		   /* 
		    * MUST unlock the closure before the queue;
		    * see rule D in the file PROTOCOLS
		    */
	  steal_give_up:
		   Closure_unlock(cl);
		   deque_unlock(victim);
		   break;

	      default:
		   Cilk_die("Bug: unknown closure status\n");
	  }

     } else {
          deque_unlock(victim);
	  Cilk_event(EVENT_STEAL_EMPTY_DEQUE);
     }

     return res;
}

/********************************************************
 * Functions that take care of remote argument sending.
 ********************************************************/

/* signal_abort_from_inlet:
 *
 * Signal all closures that need to be aborted from the parent <cl>.
 * If the parent is running and not working on the top of the 
 * stackframe, then, the current work is a spawn and should be aborted
 * as well.  In this case, we promote the child frame to a closure,
 * detach the parent, put the parent on the ready deque, and abort from
 * the parent.
 */
static void signal_abort_from_inlet(Closure *cl)
{
     Closure *child;
     Closure_assert_ownership(cl);

     if (cl->status == CLOSURE_RUNNING &&
	 !Closure_at_top_of_stack(cl)) {
	  /* 
	   * if closure is running and the frame is not on the top,
	   * then, poll_inlets is called in exception_handler.  So,
	   * the worker "Self" should own ready deque.  
	   */
	  deque_assert_ownership(Self);

	  /* promote the child frame, and detach the parent from deque */
	  child = promote_child(cl, Self);
	  finish_promote(cl, child);
	  
	  /*
	   * Now, the stackframe the exception handler is working on
	   * belongs to the child, not the parent.  Therefore, the
	   * exception handler should get the child from the deque
	   * after calling poll_inlets 
	   */
     }

     /* do NOT abort the frame at the top of the stack */
     Cilk_enter_state(STATE_ABORT_RECURSIVE);
     recursively_signal_abort(cl, ALMOST_NO_ABORT);
     Cilk_exit_state(STATE_ABORT_RECURSIVE);
}

/*
 * apply_inlet:
 *
 * Call inlet function only if the closure is not being aborted.  Use
 * abort_flag to check if abort happens when inlet function is
 * executed.  If abort occurs, signal all closures that need to be
 * aborted.
 */
static void apply_inlet(Closure *cl, struct InletClosure *i)
{
     int index = i->index;
     int argsize = i->argsize;
     void (*inlet) (void *, void *, void *) = i->inlet;
     StackFrame *frame = cl->frame;
     void *inlet_args = i->inlet_args;

     Closure_assert_ownership(cl);

     /* 
      * do not apply inlets to aborting closures --- these inlets
      * are supposed to be killed
      */
     if (!Closure_being_aborted(cl)) {
	  USE_PRIVATE(abort_flag) = 0;

	  if (inlet) {
	       /* run the inlet */
	       inlet(frame, inlet_args, (char *) frame + index);
	  } else {
	       /* 
		* default inlet: copy just-returned value into frame 
		*/
	       if (argsize) {
		    CILK_ASSERT(index >= sizeof(StackFrame));
		    memcpy((char *) frame + index, inlet_args, argsize);
	       }

	       /* 
		* TODO: make a little stub that does the memcpy.
		* (Probably the resulting code will be messier than
		*  just doing the check here)
		*/
	  }

	  if (USE_PRIVATE(abort_flag)) 
	       signal_abort_from_inlet(cl);
     }

     /* free the arguments to the inlet, and the inlet structure itself */
     if (inlet_args && argsize)
	  Cilk_internal_free(inlet_args, argsize);
     Cilk_internal_free(i, sizeof(struct InletClosure));
}

/* walk the list of inlets and call apply_inlets on all of them */
static void poll_inlets(Closure *cl)
{
     struct InletClosure *i;

     Closure_assert_ownership(cl);
     if (cl->status == CLOSURE_RUNNING &&
	 !Closure_at_top_of_stack(cl))
	  /* 
	   * If we get here, poll_inlets has been called by
	   * the exception handler 
	   */
	  deque_assert_ownership(Self);

     Cilk_event(EVENT_POLL_INLETS);

     for (i = cl->complete_inlets; i; i = cl->complete_inlets) {
	  cl->complete_inlets = i->next;

	  /*
	   * apply_inlets may change the ready deque. See comments in
	   * apply_inlet for details 
	   */
	  apply_inlet(cl, i); 
     }
     
     /* 
      * Reset the abort status after all incoming inlets have been
      * executed/killed
      */
     if (CLOSURE_ABORT(cl) == ALMOST_NO_ABORT)
	  reset_abort(cl);
}

static void complete_and_enque_inlet(Closure *parent, Closure *child, 
				     struct InletClosure *i)
{
     Closure_assert_ownership(parent);
     Cilk_event(EVENT_RETURN_ENQUEUE);

     /*
      * Install value child is returning into inlet structure.
      */
     CILK_COMPLAIN(child->return_size <= i->argsize, 
         ("Cilk runtime system: invalid size of a return-ed value.\n"
	  "Either some internal Cilk data structure is corrupted,\n"
	  "or you have inconsistent prototypes across different files.\n"
	  "(E.g., cilk int foo(...)  and  cilk double foo(...))\n"));

     if (child->return_size) {
          CILK_ASSERT(i->inlet_args != NULL);
          memcpy(i->inlet_args, child->return_value, child->return_size);
	  Cilk_internal_free(child->return_value, child->return_size);
	  WHEN_CILK_DEBUG(child->return_value = NULL);
	  WHEN_CILK_DEBUG(child->return_size = 0);
     }

     if (Closure_being_aborted(child)) {
	  if (i->inlet_args && i->argsize)
	       Cilk_internal_free(i->inlet_args, i->argsize);
	  Cilk_internal_free(i, sizeof(struct InletClosure));
     } else {
	  i->next = parent->complete_inlets;
	  parent->complete_inlets = i;
     }
}

static Closure *provably_good_steal_maybe(Closure *parent)
{
     Closure *res = (Closure *) NULL;

     Closure_assert_ownership(parent);

     if (!Closure_has_children(parent) &&
	 parent->status == CLOSURE_SUSPENDED) {
	  /* do a provably-good steal; this is *really* simple */
	  res = parent;

	  poll_inlets(parent);

	  Cilk_event(EVENT_PROVABLY_GOOD_STEAL);

	  /* debugging stuff */
	  CILK_ASSERT(parent->owner_ready_deque == NOBODY);
	  Closure_make_ready(parent);
     }

     return res;
}

/*
 * Return protocol.
 *
 * This protocol deposits the return value of a subcomputation (child)
 * into the appropriate place (parent).  This function returns a
 * closure to be executed next, or NULL if none.
 * The child must not be locked by ourselves, and be in no deque.
 */
static Closure *Closure_return(Closure *child)
{
     Closure *parent;
     Closure *res = (Closure *) NULL;
     struct InletClosure *inlet;

     CILK_ASSERT(child);
     CILK_ASSERT(child->join_counter == 0);
     CILK_ASSERT(child->status == CLOSURE_RETURNING);
     CILK_ASSERT(child->owner_ready_deque == NOBODY);
     Closure_assert_alienation(child);

     parent = child->parent;
     CILK_ASSERT(parent);

     /* 
      * at this point the status is as follows: the child is in
      * no deque and unlocked.  However, the child is still
      * in the children list of the parent.  Since the child
      * is not locked, we can safely grab the parent's lock.
      */
     Closure_lock(parent);

     CILK_ASSERT(parent->status != CLOSURE_RETURNING);
     CILK_ASSERT(parent->frame);
     CILK_ASSERT(parent->frame->magic == CILK_STACKFRAME_MAGIC);

     inlet = Closure_remove_child(parent, child);

     /*
      * the lock is not really necessary, since if rule D
      * is obeyed no one can have a pointer to child at this
      * point.  But, better be paranoid...
      */
     Closure_lock(child);

     complete_and_enque_inlet(parent, child, inlet);

     /* tell the parent to poll this inlet (it may be aborting) */
     if (parent->status == CLOSURE_RUNNING)
	  signal_immediate_exception(parent);
     else {
	  /* run them yourself */
	  poll_inlets(parent);
     }

     /* update critical path and work */
     WHEN_CILK_TIMING({
	  if (USE_PRIVATE(cp_hack) > parent->cp)
	       parent->cp = USE_PRIVATE(cp_hack);
	  parent->work += USE_PRIVATE(work_hack);
     });
	
     /* 
      * the two fences ensure dag consistency (Backer)
      */
     CILK_ASSERT(parent->join_counter > 0);
     Cilk_fence();
     --parent->join_counter;
     Cilk_fence();

     res = provably_good_steal_maybe(parent);

     Closure_unlock(parent);

     /* now the child is no longer needed */
     Closure_destroy(child);

     return res;
}

/*
 * suspend protocol 
 */
static void Closure_suspend(Closure *cl)
{
     Closure *cl1;

     Closure_checkmagic(cl);
     Closure_assert_ownership(cl);
     deque_assert_ownership(Self);

     CILK_ASSERT(cl->status == CLOSURE_RUNNING);
     cl->status = CLOSURE_SUSPENDED;

     Cilk_event(EVENT_SUSPEND);
     cl1 = deque_xtract_bottom(Self);

     CILK_ASSERT(cl == cl1);
}

/*************************************************************
 * Various interactions with the cilk2c output
 *************************************************************/

/* at a slow sync; return 0 if the sync succeeds, and 1 if suspended */
int Cilk_sync(void)
{
     Closure *t;
     int res = 0;

     Cilk_event(EVENT_CILK_SYNC);

     deque_lock(Self);

     t = deque_peek_bottom(Self);

     Closure_lock(t);
     CILK_ASSERT(t->status == CLOSURE_RUNNING);
     CILK_ASSERT(t->frame->magic == CILK_STACKFRAME_MAGIC);

     /* assert we are really at the top of the stack */
     CILK_ASSERT(Closure_at_top_of_stack(t));

     poll_inlets(t);
     deque_assert_is_bottom(t);

     maybe_reset_abort(t);

     if (Closure_has_children(t)) {
	  Closure_suspend(t);
	  res = 1;
     } 

     Closure_unlock(t);

     deque_unlock(Self);
     return res;
}

#if CILK_TIMING
/* update work and CP after a slow sync */
void Cilk_after_sync_slow_cp(Cilk_time * work, Cilk_time * cp)
{
     Closure *t;

     deque_lock(Self);
     t = deque_peek_bottom(Self);

     Closure_lock(t);

     CILK_ASSERT(t->status == CLOSURE_RUNNING);
     CILK_ASSERT(t->frame->magic == CILK_STACKFRAME_MAGIC);

     /* assert we are really at the top of the stack */
     CILK_ASSERT(Closure_at_top_of_stack(t));

     *work += t->work;
     if (t->cp > *cp)
	  *cp = t->cp;
     t->cp = 0;
     t->work = 0;

     Closure_unlock(t);
     deque_unlock(Self);
}
#endif

static void move_result_into_closure(Closure *t, void *resultp,
				     int size)
{
     Closure_assert_ownership(t);
     t->return_size = size;

     if (size) {
          t->return_value = Cilk_internal_malloc(size);
	  memcpy (t->return_value, (void *)resultp, size);
     } else {
	  WHEN_CILK_DEBUG(t->return_value = NULL);
     }
}

/*
 * This is the exception handler.
 *
 * The handler returns 1 if the worker must return to the runtime 
 * system, 0 otherwise.  This is what it does:
 *  
 * We need to call poll_inlets first. 
 * But poll_inlets may change the closure that owns the stack that 
 * is currently running.  So, we need to get the closure from the 
 * ready deque.  Then, we follow the protocol below:

 *   reset exception pointer
 *   if ( H>=T )
 *       it is a steal
 *       change state into CLOSRUE_RETURNING
 *       res = 1  
 *
 *   else if (current frame is being aborted)
 *       res = 1
 *       signal another exception for parent frame
 *       if (current frame is the last frame)
 *            do a sync
 *            {suspend if needed; o.w. destroy frame and return}
 *       else
 *            destroy current frame
 *
 *   else { not steal, not abort}
 *       reset abort pointer to NO_ABORT
 *
 *   return res;
 */
int Cilk_exception_handler(void *resultp, int size)
{
     Closure *t, *t1;
     int res = 0;

     Cilk_event(EVENT_EXCEPTION);

     deque_lock(Self);
     t = deque_peek_bottom(Self);

     CILK_ASSERT(t);
     Closure_lock(t);
     poll_inlets(t);

     t1 = deque_peek_bottom(Self);
     CILK_ASSERT(t1);

     /*
      * poll_inlets may change the bottom of the ready deque.
      * (see comments at signal_abort_from_inlet)
      */
     if (t != t1) {
	  Closure_unlock(t);
	  t = t1;
	  Closure_lock(t);
     }

     reset_exception_pointer(t);

     CILK_ASSERT(t->status == CLOSURE_RUNNING ||
		 t->status == CLOSURE_RETURNING);

     if (CLOSURE_HEAD(t) >= CLOSURE_TAIL(t)) {
	  Cilk_event(EVENT_EXCEPTION_STEAL);
	  res = 1;

	  if (t->status == CLOSURE_RUNNING) {
	       t->status = CLOSURE_RETURNING;
	       move_result_into_closure(t, resultp, size);
	  }
     } else {
	  CILK_ASSERT(t->frame->magic == CILK_STACKFRAME_MAGIC);
	  if (Closure_being_aborted(t)) {
	       StackFrame *f;
	       res = 1;

	       Cilk_event(EVENT_EXCEPTION_ABORT);
	       signal_immediate_exception(t);

	       if (Closure_at_top_of_stack(t)) {
		    if (Closure_has_children(t) != 0) {
			 Cilk_event(EVENT_EXCEPTION_ABORT_SUSPEND);
			 /* pretend to sync */
			 maybe_reset_abort(t);
			 Closure_suspend(t);
			 goto skip_free;
		    } else {
			 /* pretend we are returning */
			 t->status = CLOSURE_RETURNING;
			 Cilk_event(EVENT_EXCEPTION_ABORT_RETURN);
		    }
	       }

	       f = (StackFrame *) CLOSURE_TAIL(t)[-1];
	       Cilk_destroy_frame(f, get_frame_size(f));
	  skip_free: ;
	  } else {
	       /* frame not being aborted */
	       maybe_reset_abort(t);
	       Cilk_event(EVENT_EXCEPTION_OTHER);
	  }
     }

     Closure_unlock(t);
     deque_unlock(Self);
     return res;
}

/*
 * before a return in the slow code.  This function marks the
 * closure as returning, so that it won't be stolen.
 */
void Cilk_set_result(void *resultp, int size)
{
     Closure *t;

     Cilk_event(EVENT_RETURN_SLOW);

     deque_lock(Self);
     t = deque_peek_bottom(Self);
     Closure_lock(t);

     CILK_ASSERT(t->status == CLOSURE_RUNNING);
     CILK_ASSERT(t->frame->magic == CILK_STACKFRAME_MAGIC);

     /* 
      * this must happen after a sync; hence, either there is
      * no abort at all, or the parent is trying to kill us
      */
     CILK_ASSERT(CLOSURE_ABORT(t) == NO_ABORT ||
		 CLOSURE_ABORT(t) == ABORT_ALL);

     t->status = CLOSURE_RETURNING;
     t->frame = (StackFrame *) NULL;

     move_result_into_closure(t, resultp, size);

     Closure_unlock(t);
     deque_unlock(Self);
}


void Cilk_abort_slow(void) 
{
     Closure *cl;
     deque_lock(Self);
     cl = deque_peek_bottom(Self);
     Closure_lock(cl);

     CILK_ASSERT(cl->status == CLOSURE_RUNNING);
     CILK_ASSERT(Closure_at_top_of_stack(cl));
     Cilk_event(EVENT_ABORT_SLOW);

     Cilk_enter_state(STATE_ABORT_RECURSIVE);
     recursively_signal_abort(cl, ALMOST_NO_ABORT);
     Cilk_exit_state(STATE_ABORT_RECURSIVE);

     poll_inlets(cl);

     Closure_unlock(cl);
     deque_unlock(Self);
}

void Cilk_abort_standalone(void) 
{
     Cilk_event(EVENT_ABORT_STANDALONE);
     USE_PRIVATE(abort_flag) = 1;
}
  
/*********************************************************
 * Execution
 *********************************************************/
static Closure *setup_for_execution(Closure *t)
{
     Closure_assert_ownership(t);

     CILK_ASSERT(t->frame->magic == CILK_STACKFRAME_MAGIC);

     t->status = CLOSURE_RUNNING;
     t->cache = USE_PRIVATE(cache) = cache_get_addr(Self);

     CLOSURE_STACK(t)[0] = t->frame;	/* push the first frame on the stack */
     CLOSURE_HEAD(t) = (volatile StackFrame**)CLOSURE_STACK(t);
     CLOSURE_TAIL(t) = (volatile StackFrame**)CLOSURE_STACK(t)+1;
     reset_exception_pointer(t);
     reset_abort(t);

     return t;
}

static Closure *return_value(Closure *t)
{
     Closure *res = NULL;

     Cilk_enter_state(STATE_RETURNING);

     CILK_ASSERT(t->status == CLOSURE_RETURNING);

     res = Closure_return(t);

     Cilk_exit_state(STATE_RETURNING);
     return res;
}

/*
 * execute the closure. If the closure is returning, take care of the
 * result.
 */
static Closure *do_what_it_says(Closure *t)
{
     Closure *res = NULL;
     StackFrame *f;

     /* t must not be locked */
     Closure_assert_alienation(t);
     Closure_lock(t);

     switch (t->status) {
	 case CLOSURE_READY:
	      if (Closure_being_aborted(t)) {
		   res = abort_ready_closure(t);
		   Closure_unlock(t);
	      } else {
		   /* just execute it */
		   setup_for_execution(t);
		   f = t->frame;
		   CILK_ASSERT(f);
		   CILK_ASSERT(f->sig);
		   CILK_ASSERT(f->magic == CILK_STACKFRAME_MAGIC);
		   poll_inlets(t);
		   Closure_unlock(t);
		   /* 
		    * MUST unlock the closure before locking the queue 
		    * (rule A in file PROTOCOLS)
		    */
		   deque_lock(Self);
		   deque_add_bottom(t, Self);
		   deque_unlock(Self);

		   /* now execute it */
		   Cilk_enter_state(STATE_WORKING);
		   (get_proc_slow(f->sig)) (f);
		   Cilk_exit_state(STATE_WORKING);
	      }

	      break;

	 case CLOSURE_RETURNING:
	      /*
	       * the return protocol assumes t is not locked,
	       * and everybody will respect the fact that t is
	       * returning
	       */
	      Closure_unlock(t);
	      res = return_value(t);
	      break;

	 default:
	      Cilk_die("BUG in do_what_it_says(), t->status = %d\n", t->status);
	      break;
     }

     return res;
}

/*********************************************************
 * The scheduler itself
 *********************************************************/
void Cilk_scheduler(Closure *t)
{
     /* 
      * t contains 'the next thing to do'.  Initially, the
      * scheduler on proc 0 executes the main closure.
      */
     int victim;

     Cilk_internal_malloc_per_processor_init();
     CILK_ASSERT(Self >= 0);
     rts_srand(Self * 162347);

#ifdef CILK_LIBRARY_VERSION
  CILK_ASSERT(t == NULL);
  while (!USE_SHARED(exit)) {

     Cilk_library_wait_for_wakeup();

     CILK_ASSERT(t == NULL);
     /* Try to steal a closure from Self to bootstrap library call into
      * scheduler loop.  (Particularly important for -nproc 1)
      */
     /*t = Closure_steal(Self);*/
#endif
     
     Cilk_enter_state(STATE_TOTAL);

     while (!USE_SHARED(done)) {
	  if (!t) {
	       /* try to get work from our local queue */
	       deque_lock(Self);
	       t = deque_xtract_bottom(Self);
	       deque_unlock(Self);
	  }

	  while (!t && !USE_SHARED(done)) {
	       /* otherwise, steal */
	       Cilk_enter_state(STATE_STEALING);
	       victim = rts_rand() % USE_SHARED_RO(active_size);
	       if (victim != Self) {
		    t = Closure_steal(victim);
		    if (!t && USE_SHARED_RO(yield_slice)
			&& !USE_SHARED(done)) {
		      Cilk_lower_priority();
		    }
	       }
	       /* Cilk_fence(); */
	       Cilk_exit_state(STATE_STEALING);
	  }

	  if (USE_SHARED_RO(yield_slice))
	       Cilk_raise_priority();

	  if (!USE_SHARED(done))
	       t = do_what_it_says(t);

	  /* 
	   * if provably-good steals happened, t will contain
	   * the next closure to execute
	   */
     }

     Cilk_exit_state(STATE_TOTAL);

#ifdef CILK_LIBRARY_VERSION
     Cilk_library_go_to_sleep();
  }
#endif

     return;
}

/*
 * initialization of the scheduler. 
 */
void Cilk_scheduler_init(void)
{
     USE_SHARED(retval) = 0;
     USE_SHARED(done) = 0;
#ifdef CILK_LIBRARY_VERSION
     USE_SHARED(exit) = 0;
#endif
     CILK_CHECK(USE_SHARED_RO(active_size) > 0,
		("Partition size must be positive\n"));
     init_deques();
     init_caches();
     Cilk_internal_malloc_global_init();
}

/* 
 * From the frame of <cl>, which should be the invoke_main closure,
 * get the inlet that should be executed when cilk_main returns.
 * (All the inlet does, is to set the proper return value and call
 * abort). 
 * Pass <res>, the exit value, to the inlet, and enque it as 
 * a complete inlet.
 */
static void make_exit_inlet_closure(Closure *cl, int res)
{
     struct InletClosure *p;

     p = make_incomplete_inlet_closure(cl, 1);

     memcpy(p->inlet_args, &res, sizeof(int));     
     p->next = cl->complete_inlets;
     cl->complete_inlets = p;
     
}

/* 
 * This procedure is called by Cilk_exit().
 * <cl> is the invoke_main closure, and <res> is the return
 * value passed in by Cilk_exit.  
 * 
 * The real work is done by first enqueing an inlet that calls
 * abort, and then either signal exception or polls the inlet.
 * So, cilk_main, which is spawned by invoke_main will be 
 * aborted.  
 * (Refer to "invoke_main.c" for details.)
 */
void Cilk_exit_from_user_main(Closure *cl, int res)
{
     Closure_assert_alienation(cl);

     Closure_lock(cl);
     make_exit_inlet_closure(cl, res);
     
     if (cl->status == CLOSURE_RUNNING)
	  signal_immediate_exception(cl);
     else
	  poll_inlets(cl);
     
     Closure_unlock(cl);
}

